# SPDX-License-Identifier: MIT
# Copyright (C) 2024-2026, Advanced Micro Devices, Inc. All rights reserved.
from dataclasses import dataclass
import os
import sys

this_dir = os.path.dirname(os.path.abspath(__file__))
AITER_CORE_DIR = os.path.abspath(f"{this_dir}/../../../")
if os.path.exists(os.path.join(AITER_CORE_DIR, "aiter_meta")):
    AITER_CORE_DIR = os.path.join(AITER_CORE_DIR, "aiter/jit/utils")  # pip install mode
else:
    AITER_CORE_DIR = os.path.abspath(
        f"{this_dir}/../../aiter/jit/utils"
    )  # develop mode
sys.path.insert(0, AITER_CORE_DIR)

from chip_info import get_gfx  # noqa: E402


@dataclass
class kernelInstanceGEMM1:
    BLOCK_SIZE: int
    MPerBlock: int
    NPerBlock: int
    KPerBlock: int
    MWaves: int
    NWaves: int
    GemmPipelineVersion: int
    Nswizzle: bool = False
    MulRoutedWeight: bool = False
    ActOP: bool = False
    CDEElementOp: str = "TypeCast"
    QuantType: int = 1
    stage: int = 1
    Adtype: str = ""
    Bdtype: str = ""
    Cdtype: str = ""

    @property
    def name(self) -> str:
        return ("_").join(
            [
                f"moe_ck2stages_gemm{self.stage}",
                ("x").join(
                    map(
                        lambda x: str(x),
                        [
                            self.BLOCK_SIZE,
                            self.MPerBlock,
                            self.NPerBlock,
                            self.KPerBlock,
                        ],
                    )
                ),
                ("x").join(map(lambda x: str(x), [self.MWaves, self.NWaves])),
                self.CDEElementOp,
                f"v{self.GemmPipelineVersion}",
                "Nswizzle" + str(int(self.Nswizzle)),
                "Quant" + str(self.QuantType),
                "MulRoutedWeight" + str(int(self.MulRoutedWeight)),
                "silu" if self.ActOP else "gelu",
                self.Adtype.upper(),
                self.Bdtype.upper(),
                self.Cdtype.upper(),
            ]
        )


@dataclass
class kernelInstanceGEMM2:
    BLOCK_SIZE: int
    MPerBlock: int
    NPerBlock: int
    KPerBlock: int
    MWaves: int
    NWaves: int
    GemmPipelineVersion: int
    Nswizzle: bool = False
    MulRoutedWeight: bool = True
    CDEElementOp: str = "TypeCast"
    QuantType: int = 1
    stage: int = 2

    @property
    def name(self) -> str:
        return ("_").join(
            [
                f"moe_ck2stages_gemm{self.stage}",
                ("x").join(
                    map(
                        lambda x: str(x),
                        [
                            self.BLOCK_SIZE,
                            self.MPerBlock,
                            self.NPerBlock,
                            self.KPerBlock,
                        ],
                    )
                ),
                ("x").join(map(lambda x: str(x), [self.MWaves, self.NWaves])),
                self.CDEElementOp,
                f"v{self.GemmPipelineVersion}",
                "Nswizzle" + str(int(self.Nswizzle)),
                "Quant" + str(self.QuantType),
                "MulRoutedWeight" + str(int(self.MulRoutedWeight)),
                self.Adtype.upper(),
                self.Bdtype.upper(),
                self.Cdtype.upper(),
            ]
        )


# fmt: off
# gemm1 out&AB:bf16/fp16
a16w16_gemm1_kernels_list_gfx950= {
#   id: kernel:             BLOCK_SIZE| MPerBLOCK| NPerBLOCK| KPerBLOCK| MWaves| NWaves|GemmPipelineVersion
     0: kernelInstanceGEMM1(       256,        32,        64,       128,     1,       4,        1,),
     1: kernelInstanceGEMM1(       256,        32,        64,        64,     1,       4,        1,),
     2: kernelInstanceGEMM1(       256,        64,        64,       128,     1,       4,        1,),
     3: kernelInstanceGEMM1(       256,        64,        64,        64,     1,       4,        1,),
     4: kernelInstanceGEMM1(       256,       128,        64,        64,     1,       4,        1,),

    #  5: kernelInstanceGEMM1(       256,        64,        64,       128,     1,       4,        3,),
    #  6: kernelInstanceGEMM1(       256,        64,        64,        64,     1,       4,        3,),
#      7: kernelInstanceGEMM1(       256,       128,        64,       128,     1,       4,        3,),
#      8: kernelInstanceGEMM1(       256,       128,        64,        64,     1,       4,        3,),
     9: kernelInstanceGEMM1(       256,       128,       128,        64,     1,       4,        3,),
    10: kernelInstanceGEMM1(       256,       256,       128,        64,     1,       4,        3,),
    11: kernelInstanceGEMM1(       256,       256,        64,        64,     1,       4,        1,),
}

a16w16_gemm1_kernels_list= {
#   id: kernel:             BLOCK_SIZE| MPerBLOCK| NPerBLOCK| KPerBLOCK| MWaves| NWaves|GemmPipelineVersion
     0: kernelInstanceGEMM1(       256,        32,        64,       128,     1,       4,        1,),
     1: kernelInstanceGEMM1(       256,        32,        64,        64,     1,       4,        1,),
     2: kernelInstanceGEMM1(       256,        64,        64,       128,     1,       4,        1,),
     3: kernelInstanceGEMM1(       256,        64,        64,        64,     1,       4,        1,),
     4: kernelInstanceGEMM1(       256,       128,        64,        64,     1,       4,        1,),

     5: kernelInstanceGEMM1(       256,        64,       128,       128,     1,       4,        3,),
     6: kernelInstanceGEMM1(       256,        64,       128,        64,     1,       4,        3,),
     7: kernelInstanceGEMM1(       256,       128,       128,       128,     1,       4,        3,),
     8: kernelInstanceGEMM1(       256,       128,       128,        64,     1,       4,        3,),
     9: kernelInstanceGEMM1(       256,       256,       128,        64,     1,       4,        3,),
    10: kernelInstanceGEMM1(       256,       256,        64,        64,     1,       4,        1,),
}
# gemm1 out:bf16/fp16 AB:fp8/i8
a8w8_gemm1_kernels_list_gfx950= {
     0: kernelInstanceGEMM1(       256,       32,         64,       256,     1,       4,        1,),
     1: kernelInstanceGEMM1(       256,       32,         64,       128,     1,       4,        1,),
     2: kernelInstanceGEMM1(       256,       64,         64,       256,     1,       4,        1,),
     3: kernelInstanceGEMM1(       256,       64,         64,       128,     1,       4,        1,),
     4: kernelInstanceGEMM1(       256,      128,         64,       128,     1,       4,        1,),

    #  5: kernelInstanceGEMM1(       256,        64,        64,       256,     1,       4,        3,),
    #  6: kernelInstanceGEMM1(       256,        64,        64,       128,     1,       4,        3,),
    #  7: kernelInstanceGEMM1(       256,       128,        64,       256,     1,       4,        3,),
    #  8: kernelInstanceGEMM1(       256,       128,        64,       128,     1,       4,        3,),
     9: kernelInstanceGEMM1(       256,       128,       128,       128,     1,       4,        3,),
    10: kernelInstanceGEMM1(       256,       256,       128,       128,     1,       4,        3,),
    11: kernelInstanceGEMM1(       256,       256,        64,       128,     1,       4,        1,),
}

a8w8_gemm1_kernels_list= {
    12: kernelInstanceGEMM1(       128,       16,         64,       128,     1,       2,        1,),
    11: kernelInstanceGEMM1(       256,       16,        128,       256,     1,       4,        1,),
     0: kernelInstanceGEMM1(       256,       32,         64,       256,     1,       4,        1,),
     1: kernelInstanceGEMM1(       256,       32,         64,       128,     1,       4,        1,),
     2: kernelInstanceGEMM1(       256,       64,         64,       256,     1,       4,        1,),
     3: kernelInstanceGEMM1(       256,       64,         64,       128,     1,       4,        1,),
     4: kernelInstanceGEMM1(       256,      128,         64,       128,     1,       4,        1,),
    10: kernelInstanceGEMM1(       256,      256,         64,       128,     1,       4,        1,),

     5: kernelInstanceGEMM1(       256,        64,       128,       256,     1,       4,        3,),
     6: kernelInstanceGEMM1(       256,        64,       128,       128,     1,       4,        3,),
     7: kernelInstanceGEMM1(       256,       128,       128,       256,     1,       4,        3,),
     8: kernelInstanceGEMM1(       256,       128,       128,       128,     1,       4,        3,),
     9: kernelInstanceGEMM1(       256,       256,       128,       128,     1,       4,        3,),
}
# gemm1 blockscale out:bf16/fp16 AB:fp8/i8
a8w8_gemm1_blockscale_kernels_list= {
     0: kernelInstanceGEMM1(       256,       16,        128,       256,     1,       4,        1,),
     1: kernelInstanceGEMM1(       128,       16,        128,       128,     1,       2,        1,),
     2: kernelInstanceGEMM1(       256,       32,        128,       128,     1,       4,        1,),
     3: kernelInstanceGEMM1(       256,       64,        128,       128,     1,       4,        3,),
     #2: kernelInstanceGEMM1(       256,      128,        128,       128,     1,       4,        3,),
}

# gemm1 out:bf16/fp16 A:fp8 B:win4
a8w4_gemm1_kernels_list= {
     0: kernelInstanceGEMM1(       256,       32,         64,       128,     1,       4,        1,),
     1: kernelInstanceGEMM1(       256,       64,         64,       128,     1,       4,        1,),
     2: kernelInstanceGEMM1(       256,      128,         64,       128,     1,       4,        1,),
     3: kernelInstanceGEMM1(       256,      256,         64,       128,     1,       4,        1,),
    #  3: kernelInstanceGEMM1(       256,       64,        128,       128,     1,       4,        3,),
    #  4: kernelInstanceGEMM1(       256,      128,        128,       128,     1,       4,        3,),
    #  5: kernelInstanceGEMM1(       256,      256,        128,       128,     1,       4,        3,),
}

# gemm1 out:bf16/fp16 A:mxfp4 B:mxfp4
a4w4_gemm1_kernels_list= {
     0: kernelInstanceGEMM1(       256,       32,          128,       128,     1,       4,        3,),
     1: kernelInstanceGEMM1(       256,       64,          128,       128,     1,       4,        3,),
     2: kernelInstanceGEMM1(       256,      128,          128,       128,     1,       4,        3,),
     4: kernelInstanceGEMM1(        64,       32,           32,       128,     1,       1,        3,),
    #  3: kernelInstanceGEMM1(       256,      256,         128,       128,     2,       2,        3,),
}

# bns gemm1 out:bf16/fp16 A:mxfp4 B:mxfp4
a4w4_bns_gemm1_kernels_list= {
     0: kernelInstanceGEMM1(       256,       32,         128,       128,     1,       4,        3,),
     1: kernelInstanceGEMM1(       256,       64,          64,       128,     2,       2,        3,),
     2: kernelInstanceGEMM1(       256,      128,          64,       128,     2,       2,        3,),
}

gemm1_kernels_dict = {
    "a16w16_gfx950": a16w16_gemm1_kernels_list_gfx950,
    "a16w16": a16w16_gemm1_kernels_list,
    "a8w8_gfx950": a8w8_gemm1_kernels_list_gfx950,
    "a8w8": a8w8_gemm1_kernels_list,
    "a8w8blkscale": a8w8_gemm1_blockscale_kernels_list,
    "a8w4": a8w4_gemm1_kernels_list,
    "a4w4": a4w4_gemm1_kernels_list,
    "a4w4_bns": a4w4_bns_gemm1_kernels_list,
}


a16w16_gemm2_kernels_list_gfx950= {
#   id: kernel:             BLOCK_SIZE| MPerBLOCK| NPerBLOCK| KPerBLOCK| MWaves| NWaves|GemmPipelineVersion
# gemm2 out&AB:bf16/fp16
     0: kernelInstanceGEMM2(       256,        32,       128,       128,     1,       4,         1,),
     1: kernelInstanceGEMM2(       256,        64,       128,       128,     1,       4,         1,),
     2: kernelInstanceGEMM2(       256,       128,       128,        64,     1,       4,         1,),
     3: kernelInstanceGEMM2(       256,       256,       128,        64,     1,       4,         1,),
    #  4: kernelInstanceGEMM2(       256,        64,       128,       128,     1,       4,         3,),
     5: kernelInstanceGEMM2(       256,       128,       128,        64,     1,       4,         3,),
     6: kernelInstanceGEMM2(       256,       256,       128,        64,     1,       4,         3,),
     7: kernelInstanceGEMM2(       256,       128,        64,        64,     1,       4,         3,),
     8: kernelInstanceGEMM2(       256,       256,        64,        64,     1,       4,         3,),
     9: kernelInstanceGEMM2(       256,        32,        64,        64,     1,       4,         1,),
    10: kernelInstanceGEMM2(       256,        64,        64,        64,     1,       4,         1,),
    11: kernelInstanceGEMM2(       256,       128,        64,        64,     1,       4,         1,),
    12: kernelInstanceGEMM2(       256,        64,       128,        64,     1,       4,         1,),
}

a16w16_gemm2_kernels_list= {
#   id: kernel:             BLOCK_SIZE| MPerBLOCK| NPerBLOCK| KPerBLOCK| MWaves| NWaves|GemmPipelineVersion
# gemm2 out&AB:bf16/fp16
     0: kernelInstanceGEMM2(       256,        32,        64,       128,     1,       4,         1,),
     1: kernelInstanceGEMM2(       256,        64,        64,       128,     1,       4,         1,),
     2: kernelInstanceGEMM2(       256,       128,        64,        64,     1,       4,         1,),
     3: kernelInstanceGEMM2(       256,       256,        64,        64,     1,       4,         1,),
     4: kernelInstanceGEMM2(       256,        64,       128,       128,     1,       4,         3,),
     5: kernelInstanceGEMM2(       256,       128,       128,        64,     1,       4,         3,),
     6: kernelInstanceGEMM2(       256,       256,       128,        64,     1,       4,         3,),
     7: kernelInstanceGEMM2(       256,        32,        64,        64,     1,       4,         1,),
     8: kernelInstanceGEMM2(       256,        64,       128,        64,     1,       4,         3,),
     9: kernelInstanceGEMM2(       256,        64,        64,        64,     1,       4,         1,),
    10: kernelInstanceGEMM2(       256,       128,        64,        64,     1,       4,         3,),
    11: kernelInstanceGEMM2(       256,       256,        64,        64,     1,       4,         3,),
}

# gemm2 out:bf16/fp16 AB:fp8/i8
a8w8_gemm2_kernels_list_gfx950= {
     0: kernelInstanceGEMM2(       256,        32,       128,       256,     1,       4,         1,),
     1: kernelInstanceGEMM2(       256,        64,       128,       256,     1,       4,         1,),
     2: kernelInstanceGEMM2(       256,       128,       128,       128,     1,       4,         1,),
     3: kernelInstanceGEMM2(       256,       256,       128,       128,     1,       4,         1,),
    #  4: kernelInstanceGEMM2(       256,        64,       128,       256,     1,       4,         3,),
     5: kernelInstanceGEMM2(       256,       128,       128,       128,     1,       4,         3,),
     6: kernelInstanceGEMM2(       256,       256,       128,       128,     1,       4,         3,),
    # inter_dim=192 default instances with KPerBlock=64
    # disabled currently due to gfx950 fp8 mfma instruction used in CK not supporting these cases
    #  10: kernelInstanceGEMM2(       256,        32,        64,       64,     1,       4,         1,),
    #  11: kernelInstanceGEMM2(       256,        64,       128,       64,     1,       4,         3,),
    #  12: kernelInstanceGEMM2(       256,       128,       128,       64,     1,       4,         3,),
    #  13: kernelInstanceGEMM2(       256,       256,       128,       64,     1,       4,         3,),
}

a8w8_gemm2_kernels_list= {
    17: kernelInstanceGEMM2(       256,        16,       128,       256,     1,       4,         1,),
    18: kernelInstanceGEMM2(       128,        16,       128,       128,     1,       2,         1,),
     0: kernelInstanceGEMM2(       256,        32,        64,       256,     1,       4,         1,),
     1: kernelInstanceGEMM2(       256,        64,        64,       256,     1,       4,         1,),
     2: kernelInstanceGEMM2(       256,       128,        64,       128,     1,       4,         1,),
     3: kernelInstanceGEMM2(       256,       256,        64,       128,     1,       4,         1,),
     4: kernelInstanceGEMM2(       256,        64,       128,       256,     1,       4,         3,),
     5: kernelInstanceGEMM2(       256,       128,       128,       128,     1,       4,         3,),
     6: kernelInstanceGEMM2(       256,       256,       128,       128,     1,       4,         3,),
     7: kernelInstanceGEMM2(       256,        32,        64,       128,     1,       4,         1,),
     8: kernelInstanceGEMM2(       256,        64,       128,       128,     1,       4,         3,),
     # inter_dim=192 default instances with KPerBlock=64
     10: kernelInstanceGEMM2(       256,        32,        64,       64,     1,       4,         1,),
     11: kernelInstanceGEMM2(       256,        64,       128,       64,     1,       4,         3,),
     12: kernelInstanceGEMM2(       256,        64,        64,       64,     1,       4,         1,),
     13: kernelInstanceGEMM2(       256,       128,       128,       64,     1,       4,         3,),
     14: kernelInstanceGEMM2(       256,       128,        64,       64,     1,       4,         3,),
     15: kernelInstanceGEMM2(       256,       256,       128,       64,     1,       4,         3,),
     16: kernelInstanceGEMM2(       256,       256,        64,       64,     1,       4,         3,),
}

# gemm2 MXDLPerWave out:bf16/fp16 AB:fp8/i8
a8w8_gemm2_blockscale_kernels_list= {
     0: kernelInstanceGEMM2(       256,       16,        128,       256,     1,       4,        1,),
     1: kernelInstanceGEMM2(       128,       16,        128,       128,     1,       2,        1,),
     2: kernelInstanceGEMM2(       256,       32,        128,       128,     1,       4,        1,),
     3: kernelInstanceGEMM2(       256,       64,        128,       128,     1,       4,        3,),
     #2: kernelInstanceGEMM2(       256,      128,        128,       128,     2,       2,        3,),
}

# gemm2 out:bf16/fp16 A:fp8 B:in4
a8w4_gemm2_kernels_list= {
     0: kernelInstanceGEMM2(       256,       32,        128,       128,     1,       4,         1,),
     1: kernelInstanceGEMM2(       256,       64,        128,       128,     1,       4,         1,),
     2: kernelInstanceGEMM2(       256,      128,        128,       128,     1,       4,         1,),
     3: kernelInstanceGEMM2(       256,      256,        128,       128,     1,       4,         1,),
    #  3: kernelInstanceGEMM2(       256,       64,        128,       128,     1,       4,         3,),
    #  4: kernelInstanceGEMM2(       256,      128,        128,       128,     1,       4,         3,),
    #  5: kernelInstanceGEMM2(       256,      256,        128,       128,     1,       4,         3,),
}
# gemm2 out:bf16/fp16 A:fp8 B:in4
a4w4_gemm2_kernels_list= {
     0: kernelInstanceGEMM2(       256,        32,        128,       128,     1,       4,         3,),
     1: kernelInstanceGEMM2(       256,        64,        128,       128,     1,       4,         3,),
     2: kernelInstanceGEMM2(       256,       128,        128,       128,     1,       4,         3,),
     4: kernelInstanceGEMM2(        64,        32,         32,       128,     1,       1,         1,),
     5: kernelInstanceGEMM2(        64,        64,         128,       128,     1,       1,         3,),
     6: kernelInstanceGEMM2(        64,       128,        128,       128,     1,       1,         3,),
    #  7: kernelInstanceGEMM2(      256,       256,         64,       128,     2,       2,         3,),
}
# gemm2 out:bf16/fp16 A:fp8 B:in4
a4w4_bns_gemm2_kernels_list= {
     0: kernelInstanceGEMM2(       64,        32,         32,       128,     1,       1,         1,),
     1: kernelInstanceGEMM2(       64,        64,         64,       128,     1,       1,         1,),
     2: kernelInstanceGEMM2(       64,       128,        128,       128,     1,       1,         1,),
     4: kernelInstanceGEMM2(      256,        32,        128,       128,     1,       4,         3,),
     5: kernelInstanceGEMM2(      256,        64,         64,       128,     2,       2,         3,),
     6: kernelInstanceGEMM2(      256,       128,         64,       128,     2,       2,         3,),
}

# fmt: on
gemm2_kernels_dict = {
    "a16w16_gfx950": a16w16_gemm2_kernels_list_gfx950,
    "a16w16": a16w16_gemm2_kernels_list,
    "a8w8_gfx950": a8w8_gemm2_kernels_list_gfx950,
    "a8w8": a8w8_gemm2_kernels_list,
    "a8w8blkscale": a8w8_gemm2_blockscale_kernels_list,
    "a8w4": a8w4_gemm2_kernels_list,
    "a4w4": a4w4_gemm2_kernels_list,
    "a4w4_bns": a4w4_bns_gemm2_kernels_list,
}


bit8_list = ["F8", "I8", "f8", "i8"]
bit16_list = ["B16", "F16", "b16", "f16"]
bit4_list = ["I4", "i4", "FP4X2", "fp4x2"]
QuantType_list = [3, 4]


def get_gemm1_kernels_list(
    Adtype: str,
    Bdtype: str,
    Cdtype: str,
    Nswizzle: bool,
    QuantType: str,
    ActOP: str,
    MulRoutedWeight: bool,
    preshuffle: bool = False,
    splitk: bool = False,
) -> list:
    arch = get_gfx()
    if Adtype in bit16_list and Bdtype in bit16_list and Adtype == Adtype:
        if arch == "gfx950":
            tag = "a16w16_gfx950"
        else:
            tag = "a16w16"
    elif (
        Adtype in bit8_list
        and Bdtype in bit8_list
        and Adtype == Adtype
        and QuantType in QuantType_list
    ):
        tag = "a8w8blkscale"
    elif Adtype in bit8_list and Bdtype in bit8_list and Adtype == Adtype:
        if arch == "gfx950":
            tag = "a8w8_gfx950"
        else:
            tag = "a8w8"
    elif (
        Adtype in bit8_list
        and Bdtype in bit4_list
        and (Adtype == "F8" or Adtype == "f8")
    ):
        tag = "a8w4"
    elif Adtype in bit4_list and Bdtype in bit4_list:
        if preshuffle:
            tag = "a4w4"
        else:
            tag = "a4w4_bns"

    else:
        raise ValueError(f"Unsupported data type combination: {Adtype}, {Bdtype}")
    kernels_list = gemm1_kernels_dict[tag]
    for id, kernel in kernels_list.items():
        kernel.MulRoutedWeight = MulRoutedWeight
        kernel.ActOP = ActOP == "silu"
        kernel.Nswizzle = Nswizzle
        kernel.QuantType = QuantType
        kernel.Adtype = Adtype
        kernel.Bdtype = Bdtype
        kernel.Cdtype = Cdtype

        if tag == "a8w4":
            kernel.CDEElementOp = "MulABScaleWint4"
        elif tag == "a8w8blkscale":
            if splitk:
                kernel.CDEElementOp = "MulABScaleExpertWeightA8W8blkscaleSplitk"
            else:
                kernel.CDEElementOp = "MulABScaleExpertWeightA8W8blkscale"
        elif tag == "a8w8" or tag == "a4w4_bns":
            kernel.CDEElementOp = "MulABScale"
        elif tag == "a4w4":
            kernel.CDEElementOp = "MulABScaleShuffled"
        elif tag == "a16w16":
            if MulRoutedWeight:
                kernel.CDEElementOp = "TypeCastExpertWeight"
            else:
                kernel.CDEElementOp = "TypeCast"
    return tag, kernels_list


def get_gemm2_kernels_list(
    Adtype: str,
    Bdtype: str,
    Cdtype: str,
    Nswizzle: bool,
    QuantType: str,
    MulRoutedWeight: bool,
    preshuffle: bool = False,
) -> list:
    arch = get_gfx()

    if Adtype in bit16_list and Bdtype in bit16_list and Adtype == Adtype:
        if arch == "gfx950":
            tag = "a16w16_gfx950"
        else:
            tag = "a16w16"
    elif (
        Adtype in bit8_list
        and Bdtype in bit8_list
        and Adtype == Adtype
        and QuantType in QuantType_list
    ):
        tag = "a8w8blkscale"
    elif Adtype in bit8_list and Bdtype in bit8_list and Adtype == Adtype:
        if arch == "gfx950":
            tag = "a8w8_gfx950"
        else:
            tag = "a8w8"
    elif (
        Adtype in bit8_list
        and Bdtype in bit4_list
        and (Adtype == "F8" or Adtype == "f8")
    ):
        tag = "a8w4"
    elif Adtype in bit4_list and Bdtype in bit4_list:
        if preshuffle:
            tag = "a4w4"
        else:
            tag = "a4w4_bns"
    else:
        raise ValueError(f"Unsupported data type combination: {Adtype}, {Bdtype}")
    kernels_list = gemm2_kernels_dict[tag]
    for id, kernel in kernels_list.items():
        kernel.MulRoutedWeight = MulRoutedWeight
        kernel.Nswizzle = Nswizzle
        kernel.QuantType = QuantType
        kernel.Adtype = Adtype
        kernel.Bdtype = Bdtype
        kernel.Cdtype = Cdtype

        if tag == "a8w4":
            kernel.CDEElementOp = "MulABScaleExpertWeightWin4"
        elif tag == "a8w8blkscale":
            kernel.CDEElementOp = "MulABScaleExpertWeightA8W8blkscale"
        elif tag == "a8w8" or tag == "a4w4_bns":
            kernel.CDEElementOp = "MulABScaleExpertWeight"
        elif tag == "a4w4":
            kernel.CDEElementOp = "MulABScaleExpertWeightShuffled"
        elif tag == "a16w16":
            if MulRoutedWeight:
                kernel.CDEElementOp = "TypeCastExpertWeight"
            else:
                kernel.CDEElementOp = "TypeCast"
    return tag, kernels_list



# SPDX-License-Identifier: MIT
# Copyright (C) 2024-2026, Advanced Micro Devices, Inc. All rights reserved.

from typing import Optional
import torch
import triton
from aiter.ops.triton._triton_kernels.gemm.batched.batched_gemm_a8w8_a_per_token_group_prequant_w_per_batched_tensor_quant import (
    _batched_gemm_a8w8_a_per_token_group_prequant_w_per_batched_tensor_quant_kernel,
    _get_config,
)


def batched_gemm_a8w8_a_per_token_group_prequant_w_per_batched_tensor_quant(
    X: torch.Tensor,
    WQ: torch.Tensor,
    w_scale: torch.Tensor,
    group_size: int = 128,
    bias: Optional[torch.Tensor] = None,
    dtype: Optional[torch.dtype] = torch.bfloat16,
    splitK: Optional[int] = None,
    YQ: Optional[torch.Tensor] = None,
    transpose_bm: Optional[bool] = False,
    transpose_bm_in: Optional[bool] = False,
    config: Optional[dict] = None,
):
    """
    Computes batched 8 bit matrix multiplication Y[i] = X[i] @ W[i]^T with active activation quantization.
    X is quantized to INT8 during computation using per-token grouped quantization.
    W is pre-quantized INT8 with per-batch-element scaling.

    Args:
        X (torch.Tensor): Higher precision input batch with shape (B, M, K) or (M, B, K) if transpose_bm_in=True.
            Quantized to INT8 on-the-fly during GEMM.
        WQ (torch.Tensor): Pre-quantized INT8 weight batch with shape (B, N, K), internally transposed.
        w_scale (torch.Tensor): Per-batch scale for WQ with shape (1,).
        group_size (int): Group size for per-token grouped quantization of X. Must be power of 2.
        bias (Optional[torch.Tensor]): Bias batch with shape (B, 1, N).
        dtype (Optional[torch.dtype]): Output datatype (BF16 or FP16).
        splitK (Optional[int]): Not supported. Must be None.
        YQ (Optional[torch.Tensor]): Pre-allocated output tensor with shape (B, M, N) or (M, B, N) if transpose_bm=True.
        transpose_bm (Optional[bool]): Transpose batch and M dimensions in output.
        transpose_bm_in (Optional[bool]): Transpose batch and M dimensions in input.
        config (Optional[dict]): Kernel tuning parameters (BLOCK_SIZE_M, BLOCK_SIZE_N, GROUP_SIZE_M).

    Returns:
        torch.Tensor: Output batch with shape (B, M, N) or (M, B, N) if transpose_bm=True.
    """

    # Check constraints.
    if not transpose_bm_in:
        B = X.shape[0]
        M = X.shape[1]
    else:
        M = X.shape[0]
        B = X.shape[1]
    K = X.shape[2]
    N = WQ.shape[1]

    assert B == WQ.shape[0], "Incompatible Batch dimensions!!!"
    assert K == WQ.shape[2], "Incompatible K dimensions!!!"
    assert (
        triton.next_power_of_2(group_size) == group_size
    ), "group_size mush be power of 2"
    assert dtype in [
        torch.bfloat16,
        torch.float16,
    ], f"Output {dtype=} is currently not supported in batched_gemm_a8w8"
    assert splitK is None, "Currently, there isn't any support for splitK on Triton"

    WQ = WQ.transpose(1, 2)

    has_bias = bias is not None
    if YQ is None:
        if transpose_bm:
            YQ = torch.empty((M, B, N), dtype=dtype, device=X.device)
        else:
            YQ = torch.empty((B, M, N), dtype=dtype, device=X.device)
    else:
        if transpose_bm:
            assert (
                YQ.shape[0] == M and YQ.shape[1] == B and YQ.shape[2] == N
            ), "Output dimension error"
        else:
            assert (
                YQ.shape[0] == B and YQ.shape[1] == M and YQ.shape[2] == N
            ), "Output dimension error"

    if config is None:
        config, _ = _get_config(M, N, K)
    config["BLOCK_SIZE_K"] = group_size
    config["kpack"] = 1

    grid = lambda META: (  # noqa: E731
        B,
        triton.cdiv(M, META["BLOCK_SIZE_M"]) * triton.cdiv(N, META["BLOCK_SIZE_N"]),
    )

    DTYPE_MAX = (
        torch.finfo(WQ.dtype).max
        if torch.is_floating_point(WQ)
        else torch.iinfo(WQ.dtype).max
    )

    _batched_gemm_a8w8_a_per_token_group_prequant_w_per_batched_tensor_quant_kernel[
        grid
    ](
        X,
        WQ,
        YQ,
        w_scale,
        bias,
        M,
        N,
        K,
        X.stride(0) if not transpose_bm_in else X.stride(1),
        X.stride(1) if not transpose_bm_in else X.stride(0),
        X.stride(2),
        WQ.stride(0),
        WQ.stride(1),
        WQ.stride(2),
        YQ.stride(0) if not transpose_bm else YQ.stride(1),
        YQ.stride(1) if not transpose_bm else YQ.stride(0),
        YQ.stride(2),
        bias.stride(0) if has_bias else 0,
        has_bias,
        DTYPE_MAX=DTYPE_MAX,
        DTYPE_MIN=-DTYPE_MAX,
        **config,
    )

    return YQ



# SPDX-License-Identifier: MIT
# Copyright (C) 2023-2025, Advanced Micro Devices, Inc. All rights reserved.
# generate kernel instances to speed up compilation

import argparse
from dataclasses import dataclass
from pathlib import Path
from typing import Any, List


def get_if_str(idx, total, last_else=True):
    if idx == 0:
        return "if"
    else:
        return "else if"


DATA_TYPE_MAP = {
    "float32": "float",
    "float64": "double",
    "int32": "int",
    "int64": "long long",
    "bool": "bool",
    "float16": "torch::Half",
    "bfloat16": "torch::BFloat16",
}

TORCH_TYPE_MAP = {
    "float32": "torch::kFloat32",
    "float64": "torch::kFloat64",
    "int32": "torch::kInt32",
    "int64": "torch::kInt64",
    "bool": "torch::kBool",
    "float16": "torch::kHalf",
    "bfloat16": "torch::kBFloat16",
}

OPERATOR_MAP = {
    "add": "aiter::AddOp",
    "sub": "aiter::SubOp",
    "mul": "aiter::MulOp",
    "div": "aiter::DivOp",
}


class BinaryOpCodegen:
    API_COMMON_HEADER = """
// SPDX-License-Identifier: MIT
// Copyright (c) 2023, Advanced Micro Devices, Inc. All rights reserved.
#pragma once
#include <torch/extension.h>
#include "binary_operator.cuh"

template <typename Op, typename T0, typename T1>
void binary_op_impl(torch::Tensor &input, torch::Tensor &other, torch::Tensor &output)
{
  const at::hip::OptionalHIPGuardMasqueradingAsCUDA device_guard(device_of(input));
  int dim = input.dim();

  bool is_support = false;
  bool order_flag = true;
  int pattern = 0;
  constexpr uint32_t PATTERN_TRANSPOSE = 1;
  constexpr uint32_t PATTERN_BROADCAST_0 = 2;   // (m, n, k), (1, n, k)
  constexpr uint32_t PATTERN_BROADCAST_1 = 3;   // (m, n, k), (m, 1, k)
  constexpr uint32_t PATTERN_CONTIGUOUS = 4;
  constexpr uint32_t PATTERN_BROADCAST_2 = 5;   // (m, n, k), (m, n, 1)
  constexpr uint32_t PATTERN_BROADCAST_3 = 6;   // (m, n, k), (   n, 1)
  constexpr uint32_t PATTERN_BROADCAST_4 = 7;   // (m, n, k), (m, 1, 1)

  // contiguous case
  if (!is_support)
  {
    is_support = true;
    is_support &= (input.dim() == other.dim());
    is_support &= input.is_contiguous() == other.is_contiguous();
    is_support &= input.is_contiguous() == true;
    if (input.dim() == 1)
    {
      is_support &= input.numel() % 128 == 0;
    }
    for (int i = 0; i < input.dim() && is_support; ++i)
    {
      is_support &= (input.size(i) == other.size(i));
    }
    pattern = is_support ? PATTERN_CONTIGUOUS : 0;
  }

  if (!is_support && (dim == 3 || other.dim() == 3))
  {
    // transpose case
    if (input.is_contiguous() != other.is_contiguous())
    {
      auto tensor_not_conti = input.is_contiguous() ? other : input;
      order_flag = !input.is_contiguous() ? true : false;
      is_support = true;
      // avoid broadcast
      is_support &= input.dim() == other.dim();
      is_support &= input.size(0) == other.size(0);
      is_support &= input.size(1) == other.size(1);
      is_support &= input.size(2) == other.size(2);
      is_support &= tensor_not_conti.stride(1) == 1;
      pattern = is_support ? PATTERN_TRANSPOSE : 0;
    }
    // broadcast case
    else if (input.is_contiguous() && other.is_contiguous())
    {
      is_support = false;
      // input tensor dim and other tensor dim both equal to 3
      if (input.dim() == other.dim())
      {
        // broadcast at dim0 or dim1 or dim2
        auto broadcast_3d_case = [&] (int bcast_dim)
        {
          constexpr int bcast_pattern[3] = {PATTERN_BROADCAST_0, PATTERN_BROADCAST_1, PATTERN_BROADCAST_2};
          if (!is_support && (input.size(bcast_dim) == 1 || other.size(bcast_dim)) && input.size(bcast_dim) != other.size(bcast_dim))
          {
            is_support = true;
            for (int i = 0; i < 3; ++i)
            {
              if (bcast_dim != i) is_support &= input.size(i) == other.size(i);
            }
            is_support &= input.size(bcast_dim) == 1 ? other.size(bcast_dim) != 1 : true;
            pattern = is_support ? bcast_pattern[bcast_dim] : 0;
            order_flag = input.size(bcast_dim) != 1 ? true : false;
            // if (bcast_dim == 1) order_flag = !order_flag;
          }
        };
        // (m, n, k), (1, n, k) or (1, n, k), (m, n, k)
        broadcast_3d_case(0);
        // (m, n, k), (m, 1, k) or (m, 1, k), (m, n, k)
        broadcast_3d_case(1);
        // (m, n, k), (m, n, 1) or (m, n, 1), (m, n, k)
        broadcast_3d_case(2);


        bool first_dim_eq = input.size(0) == other.size(0);
        bool input_bcast_last2dim = input.size(1) == 1 && input.size(2) == 1;
        bool other_bcast_last2dim = other.size(1) == 1 && other.size(2) == 1;
        if (first_dim_eq && (input_bcast_last2dim || other_bcast_last2dim)) // broadcast in last 2 dim
        {
          is_support = true;
          order_flag = other_bcast_last2dim ? true : false;
          pattern = PATTERN_BROADCAST_4;
        }
      }
      // (m, n, k), (n, 1) or (n, 1), (m, n, k)
      else if (input.dim() == 2 || other.dim() == 2)
      {
        is_support = true;
        if (input.dim() == 2)
        {
          is_support &= input.size(0) == other.size(1);
          is_support &= input.size(1) == 1;
          pattern = is_support ? PATTERN_BROADCAST_3 : 0;
          order_flag = false;
        }
        else
        {
          is_support &= other.size(0) == input.size(1);
          is_support &= other.size(1) == 1;
          pattern = is_support ? PATTERN_BROADCAST_3 : 0;
          order_flag = true;
        }
      }
      // (m, n, k), (k) or (k), (m, n, k)
      // (m, n, k), (1) or (1), (m, n, k)
      else if (input.dim() == 1 || other.dim() == 1)
      {
        if (other.dim() == 1)
        {
          if (other.size(0) == 1 || (other.size(0) == input.size(2) && input.size(2) % (128 / input.element_size()) == 0))
          {
            if (input.numel() % (256 * 8 * 16 / input.element_size()) == 0)
            {
              is_support = true;
              pattern = PATTERN_BROADCAST_0;
              order_flag = true;
            }
          }
        }
        else
        {
          if (input.size(0) == 1 || (input.size(0) == other.size(2) && other.size(2) % (128 / other.element_size()) == 0))
          {
            if (other.numel() % (256 * 8 * 16 / other.element_size()) == 0)
            {
              is_support = true;
              pattern = PATTERN_BROADCAST_0;
              order_flag = false;
            }
          }
        }
      }
    }
  }

  if (!is_support && input.dim() != 3 && other.dim() != 3)
  {
    if (input.dim() == other.dim())
    {
      std::vector<int> bcast_dim_index = {};
      for (int i = 0; i < input.dim(); ++i)
      {
        // broadcast condition
        if (input.size(i) != other.size(i) && (input.size(i) == 1 || other.size(i) == 1))
        {
          bcast_dim_index.push_back(i);
        }
      }
      if (bcast_dim_index.size() == 1 && bcast_dim_index[0] != 0 && bcast_dim_index[0] != input.dim() - 1)
      {
        is_support = true;
        pattern = PATTERN_BROADCAST_1;
        order_flag = other.size(bcast_dim_index[0]) == 1 ? true : false;
      }
    }
  }

  // hip does not support double
  if (input.dtype() == torch::kDouble || other.dtype() == torch::kDouble)
  {
    is_support = false;
  }

  if (is_support)
  {
    switch (pattern)
    {
      case PATTERN_TRANSPOSE:
        binary_operation_process<1, Op, T0, T1>(input, other, output, order_flag);
        break;
      case PATTERN_BROADCAST_0:
        binary_operation_process<2, Op, T0, T1>(input, other, output, order_flag);
        break;
      case PATTERN_BROADCAST_1:
        binary_operation_process<3, Op, T0, T1>(input, other, output, order_flag);
        break;
      case PATTERN_CONTIGUOUS:
        binary_operation_process<4, Op, T0, T1>(input, other, output, order_flag);
        break;
      case PATTERN_BROADCAST_2:
        binary_operation_process<5, Op, T0, T1>(input, other, output, order_flag);
        break;
      case PATTERN_BROADCAST_3:
        binary_operation_process<6, Op, T0, T1>(input, other, output, order_flag);
        break;
      case PATTERN_BROADCAST_4:
        binary_operation_process<7, Op, T0, T1>(input, other, output, order_flag);
        break;
      default:
        return ;
    }
    return ;
  }
  else
  {
  output = aiter::aten_compute<Op>(input, other);
  return ;
  }
}
"""

    API_BASE = """
// SPDX-License-Identifier: MIT
// Copyright (c) 2023, Advanced Micro Devices, Inc. All rights reserved.

#include "binary_op_api_common.hpp"
void binary_op_dispatch(const std::string& op_type,
                       torch::Tensor &input,
                       torch::Tensor &other,
                       torch::Tensor &output) {{
    // Dispatch based on operator and input types
{F_dispatch}
    //AT_ERROR("Unsupported operator or dtype combination");
}}
"""

    API_PER_OPERATOR = """    {F_if}(op_type == "{F_op_type}") {{
{F_per_dtype}
    }}
"""

    API_PER_DTYPE = """        {F_if}(input.scalar_type() == {F_in0_type} && other.scalar_type() == {F_in1_type}) {{
            binary_op_impl<{F_op_cpp}, {F_in0_cpp}, {F_in1_cpp}>(input, other, output);
        }}
"""

    INSTANCE_BASE = """
// SPDX-License-Identifier: MIT
// Copyright (c) 2023, Advanced Micro Devices, Inc. All rights reserved.

#include "binary_op_api_common.hpp"

// Explicit instantiation
{F_instance_def}
"""

    def __init__(self, working_path, dtypes, optype):
        self.working_path = working_path
        self.dtype_cmb = []
        for i in dtypes:
            input_dtype = i.split("_")[0]
            other_dtype = i.split("_")[1]
            self.dtype_cmb.append((str(input_dtype), str(other_dtype)))
        self.op_type = optype

    @dataclass
    class h_traits:
        F_op_type: str
        F_in0_type: str
        F_in1_type: str

        @property
        def trait_name(self) -> str:
            return f"{OPERATOR_MAP[self.F_op_type]}, {DATA_TYPE_MAP[self.F_in0_type]}, {DATA_TYPE_MAP[self.F_in1_type]}"

        @property
        def call_name(self) -> str:
            return f"binary_op_impl<{self.trait_name}>"

        @property
        def def_name(self) -> str:
            return f"template void binary_op_impl<{self.trait_name}>(torch::Tensor&, torch::Tensor&, torch::Tensor&);"

    @dataclass
    class h_instance:
        F_op_type: str
        F_dtype_pair: str  # "in0_type,in1_type"
        instance_list: List[Any]

        @property
        def name(self) -> str:
            in0, in1 = self.F_dtype_pair.split(",")
            return f"binary_op_{self.F_op_type}_{in0}_{in1}"

        @property
        def content(self) -> str:
            instance_defs = "\n".join(ins.def_name for ins in self.instance_list)
            return BinaryOpCodegen.INSTANCE_BASE.format(F_instance_def=instance_defs)

    def content_api(self) -> str:
        op_dict = {}
        blobs = self.get_blobs()

        for blob in blobs:
            if blob.F_op_type not in op_dict:
                op_dict[blob.F_op_type] = {}
            if blob.F_dtype_pair not in op_dict[blob.F_op_type]:
                op_dict[blob.F_op_type][blob.F_dtype_pair] = blob

        dispatch_str = ""
        for i_op, (op_type, dtype_blobs) in enumerate(op_dict.items()):
            dtype_str = ""
            for i_dtype, (dtype_key, blob) in enumerate(dtype_blobs.items()):
                in0, in1 = dtype_key.split(",")
                dtype_str += self.API_PER_DTYPE.format(
                    F_if=get_if_str(i_dtype, len(dtype_blobs)),
                    F_in0_type=TORCH_TYPE_MAP[in0],
                    F_in1_type=TORCH_TYPE_MAP[in1],
                    F_op_cpp=OPERATOR_MAP[op_type],
                    F_in0_cpp=DATA_TYPE_MAP[in0],
                    F_in1_cpp=DATA_TYPE_MAP[in1],
                )

            dispatch_str += self.API_PER_OPERATOR.format(
                F_if=get_if_str(i_op, len(op_dict)),
                F_op_type=op_type,
                F_per_dtype=dtype_str,
            )

        return self.API_BASE.format(F_traits_define="", F_dispatch=dispatch_str)

    def get_blobs(self):
        operators = self.op_type
        dtype_combinations = self.dtype_cmb

        blobs = []
        for op in operators:
            for in0, in1 in dtype_combinations:
                traits = self.h_traits(op, in0, in1)
                dtype_key = f"{in0},{in1}"
                blobs.append(self.h_instance(op, dtype_key, [traits]))

        return blobs

    def gen_blobs(self):
        w_p = Path(self.working_path)

        # Generate API files
        (w_p / "binary_op_api.cpp").write_text(self.content_api())
        (w_p / "binary_op_api_common.hpp").write_text(self.API_COMMON_HEADER)

        # Generate kernel instance files
        blobs = self.get_blobs()
        for b in blobs:
            (w_p / f"{b.name}.cpp").write_text(b.content)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="generate", description="Generate binary operation kernels"
    )
    parser.add_argument(
        "-w",
        "--working_path",
        default="./generated",
        help="Output directory for generated files",
    )

    parser.add_argument("-o", "--optype", default="add", help="binary operator optype")

    parser.add_argument(
        "-t",
        "--dtypes",
        default="float32_float32",
        help="input tensor, other tensor dtype",
    )

    args = parser.parse_args()

    p = Path(args.working_path)
    if not p.exists():
        p.mkdir()
    optype_str = args.optype
    dtype_str = args.dtypes
    if args.optype == "all":
        optype_str = "add, sub, mul, div"
    if dtype_str == "all":
        all_type = [
            "float32",
            "bfloat16",
            "float16",
            "int32",
        ]
        tmp_str = ""
        for input_dtype in all_type:
            for other_dtype in all_type:
                tmp_str += input_dtype + "_" + other_dtype + ","
        dtype_str = tmp_str
    op_list = optype_str.split(",")
    op_list = [x.strip() for x in op_list if x.strip()]
    dtype_list = dtype_str.split(",")
    dtype_list = [x.strip() for x in dtype_list if x.strip()]
    BinaryOpCodegen(args.working_path, dtype_list, op_list).gen_blobs()



# SPDX-License-Identifier: MIT
# Copyright (C) 2024-2025, Advanced Micro Devices, Inc. All rights reserved.
import os
import sys
import hashlib
import aiter
import numpy as np
import argparse
import random
import tempfile
import torch
from jinja2 import Template
from aiter.test_common import perftest
from aiter import pertoken_quant, per_tensor_quant
from csrc.cpp_itfs.torch_utils import torch_to_c_types
import triton
import triton.language as tl

from csrc.cpp_itfs.gluon_aot_tools.compile_gluon import (
    compile_gluon_kernel,
    CompileGluonArgs,
)
from csrc.cpp_itfs.utils import (
    compile_template_op,
    AITER_CORE_DIR,
    get_default_func_name,
    run_lib,
)
import aiter.ops.triton.utils._triton.arch_info as arch_info
from aiter.ops.triton.gluon.pa_decode_gluon import (
    paged_attention_decode_v2_gluon_dot_kernel,
    paged_attention_decode_v2_gluon_large_block_dot_kernel,
    get_cdna_version,
)
from op_tests.triton_tests.test_pa_decode_gluon import (
    torch_attention_compute,
    create_kv_cache,
    quantize_kv_cache_symmetric,
    quantize_kv_cache_per_tensor,
    shuffle_value_cache_layout,
)

TORCH_TO_TL_DTYPE = {
    torch.float8_e4m3fnuz: tl.float8e4b8,
    torch.float8_e4m3fn: tl.float8e4nv,
    torch.bfloat16: tl.bfloat16,
    torch.float16: tl.float16,
}
UNIFORM_RANGE = (-1, 1)
compile_reduce_kernel_count = 0


def setup_seed(seed: int) -> None:
    """Set random seed for reproducibility."""
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True


def tensor_to_hash(tensor: torch.Tensor, algorithm: str = "md5") -> str:
    """
    Convert a PyTorch tensor to a hash value using the specified algorithm.

    Args:
        tensor (torch.Tensor): Input tensor
        algorithm (str): Hash algorithm, defaults to 'md5',
                        options: 'md5', 'sha1', 'sha256', etc.

    Returns:
        str: Hexadecimal string representation of the hash value
    """
    hash_func = getattr(hashlib, algorithm)()

    # Process tensor data
    tensor_data = tensor.contiguous().view(torch.uint8).detach().cpu().numpy().tobytes()

    hash_func.update(tensor_data)
    return hash_func.hexdigest()


def compile_ttgir_with_triton(ttgir_content: str):
    """
    Compile TTGIR (Triton Tensor IR) content to executable artifact.

    This function takes TTGIR string content, writes it to a temporary file,
    and compiles it using the Triton compiler. The temporary file is cleaned up
    after compilation regardless of success or failure.

    Args:
        ttgir_content (str): The TTGIR (Triton Tensor IR) code as a string
                             to be compiled.

    Returns:
        object: The compiled artifact from the Triton compiler.

    Raises:
        Exception: Any exception raised during the compilation process
                  will be propagated to the caller.

    Note:
        This function uses a temporary file to work with the Triton compiler
        which expects file input. The file is automatically deleted after
        compilation to avoid leaving temporary files on disk.
    """
    # Create a temporary file to store the TTGIR content
    # Using NamedTemporaryFile with delete=False to control deletion manually
    with tempfile.NamedTemporaryFile(
        mode="w", suffix=".ttgir", delete=False
    ) as temp_file:
        # Write TTGIR content to temporary file
        temp_file.write(ttgir_content)
        ttgir_file_path = temp_file.name

    try:
        # Compile the TTGIR file using Triton compiler
        # This converts the intermediate representation to executable code
        compiled_artifact = triton.compiler.compiler.compile(ttgir_file_path)
        return compiled_artifact

    finally:
        # Ensure temporary file is cleaned up even if compilation fails
        # This prevents leaving temporary files on the filesystem
        if os.path.exists(ttgir_file_path):
            os.unlink(ttgir_file_path)


def compile_attention_kernel(
    compute_type: torch.dtype,
    query_seq_len: int,
    one_query_group_size: int,
    head_size: int,
    kv_block_size: int,
    context_partition_size: int,
    query_quant_mode: int,
    kv_quant_mode: int,
    fp8_max_value: float,
    value_transposed: int,
    is_causal: int,
    cdna_version: int,
    md_name: str,
    func_name: str = None,
):
    """Compile the attention kernel for paged attention decode."""
    head_size_pow2 = triton.next_power_of_2(head_size)
    equi_query_group_size_pow2 = triton.next_power_of_2(
        query_seq_len
    ) * triton.next_power_of_2(one_query_group_size)

    if func_name is None:
        func_name = get_default_func_name(
            md_name,
            (
                compute_type,
                query_seq_len,
                one_query_group_size,
                head_size_pow2,
                kv_block_size,
                context_partition_size,
                query_quant_mode,
                kv_quant_mode,
                fp8_max_value,
                value_transposed,
                is_causal,
                cdna_version,
            ),
        )

    global compile_reduce_kernel_count
    compile_reduce_kernel_count += 1

    if compile_reduce_kernel_count == 1:
        # Convert compute_type from torch.dtype to tl.dtype for AOT compilation
        compute_type_tl = TORCH_TO_TL_DTYPE[compute_type]

        kv_compute_block_size = 256
        waves_per_eu = 1
        # Select kernel implementation based on block size
        if kv_block_size > context_partition_size:
            # Use big block kernel for large block sizes
            if value_transposed:
                # Use smaller compute block size for better performance with transposed values
                kv_compute_block_size = 128
        else:
            # Use standard kernel for normal block sizes
            # Configure waves per EU based on query group size
            if equi_query_group_size_pow2 == 64:
                waves_per_eu = 3
            else:
                waves_per_eu = 4

        if compute_type_tl == tl.float8e4b8 or compute_type_tl == tl.bfloat16:
            if query_quant_mode >= 0:
                query_sig = "*fp8e4b8:16"
            else:
                query_sig = "*bf16:16"
            if kv_quant_mode >= 0:
                key_cache_sig = "*fp8e4b8:16"
                value_cache_sig = "*fp8e4b8:16"
            else:
                key_cache_sig = "*bf16:16"
                value_cache_sig = "*bf16:16"
            logits_sig = "*bf16:16"
        elif compute_type_tl == tl.float16:
            if query_quant_mode >= 0:
                query_sig = "*fp8e4b8:16"
            else:
                query_sig = "*fp16:16"
            if kv_quant_mode >= 0:
                key_cache_sig = "*fp8e4b8:16"
                value_cache_sig = "*fp8e4b8:16"
            else:
                key_cache_sig = "*fp16:16"
                value_cache_sig = "*fp16:16"
            logits_sig = "*fp16:16"
        else:
            raise ValueError(f"Unsupported compute type: {compute_type_tl}")
        # Build signature based on kernel parameters (combined from both kernels)
        signature_parts = [
            "*fp32:16",  # exp_sums_ptr
            "*fp32:16",  # max_logits_ptr
            logits_sig,  # logits_ptr (output_ptr for attention kernel)
            query_sig,  # query_ptr [batch_size, query_length, num_kv_heads, query_group_size, head_size]
            key_cache_sig,  # key_cache_ptr
            value_cache_sig,  # value_cache_ptr
            "*i32:16",  # block_tables_ptr
            "*i32:16",  # context_lengths_ptr
            "fp32:16",  # softmax_scale
            "*fp32:16",  # query_scale [num_seqs, query_length, num_kv_heads, query_group_size, 1](per-token) or [1](per-tensor) or None
            "*fp32:16",  # key_scale
            "*fp32:16",  # value_scale
            "i32:16",  # stride_max_logits_seq
            "i32:16",  # stride_max_logits_head
            "i32:16",  # stride_max_logits_part
            "i32:16",  # stride_output_seq
            "i32:16",  # stride_output_head
            "i32:16",  # stride_output_part
            "i32:16",  # stride_output_group
            # 5D query strides for [batch_size, query_length, num_kv_heads, query_group_size, head_size]
            "i32:16",  # stride_query_bs
            "i32:16",  # stride_query_qlen
            "i32:16",  # stride_query_kv_head
            "i32:16",  # stride_query_group_size
            "i32:16",  # stride_key_block
            "i32:16",  # stride_key_head
            "i32:16",  # stride_key_head_split
            "i32:16",  # stride_key_block_elem
            "i32:16",  # stride_value_block
            "i32:16",  # stride_value_head
            "i32:16",  # stride_value_head_size
            "i32:16",  # stride_block_table_seq
            # 5D query_scale strides
            "i32:16",  # stride_query_scale_bs
            "i32:16",  # stride_query_scale_qlen
            "i32:16",  # stride_query_scale_kv_head
            "i32:16",  # kv_scale_stride_0
            "i32:16",  # kv_scale_stride_1
            "i32:16",  # head_size
            "i32:16",  # num_seqs
            "i32:16",  # num_kv_heads
            "i32:16",  # max_context_partition_num
            f"{str(compute_type_tl)}",
            f"{query_seq_len}",  # QUERY_SEQ_LEN (constexpr)
            f"{one_query_group_size}",  # ONE_QUERY_GROUP_SIZE (constexpr)
            f"{head_size_pow2}",
            f"{kv_block_size}",
            f"{context_partition_size}",
            f"{kv_compute_block_size}",
            f"{query_quant_mode}",
            f"{kv_quant_mode}",
            f"{fp8_max_value}",
            f"{value_transposed}",
            f"{is_causal}",
            f"{cdna_version}",
        ]
        signature = ",".join(signature_parts)
        gluon_kernel_name = "paged_attention_decode_v2_gluon_dot_kernel"
        if kv_block_size > context_partition_size:
            gluon_kernel_name = "paged_attention_decode_v2_gluon_large_block_dot_kernel"

        compile_args = CompileGluonArgs(
            path=f"{AITER_CORE_DIR}/aiter/ops/triton/gluon/pa_decode_gluon.py",
            kernel_name=gluon_kernel_name,
            signature=signature,
            grid="num_seqs,num_kv_heads,max_context_partition_num",
            num_warps=4,
            waves_per_eu=waves_per_eu,
            num_stages=1,
            num_ctas=1,
            kpack=1,
            out_name=md_name,
        )
        triton_kernel, output_files = compile_gluon_kernel(compile_args)
        triton_header = None
        triton_source = None
        for output_file in output_files:
            if output_file.suffix == ".h":
                triton_header = output_file
            elif output_file.suffix == ".cpp":
                triton_source = output_file

        with open(
            f"{AITER_CORE_DIR}/csrc/cpp_itfs/pa_gluon_aot/pa_decode_attention_kernel.cpp.jinja",
            "r",
        ) as f:
            src_template = Template(f.read())

        return compile_template_op(
            src_template,
            md_name,
            [triton_header],
            [triton_source],
            triton_header=triton_header,
            kernel_name=md_name,
            triton_kernel=triton_kernel,
            func_name=func_name,
        )
    else:
        return run_lib(func_name)


@perftest()
def run_compiled_attention_kernel(
    exp_sums: torch.Tensor,
    max_logits: torch.Tensor,
    temporary_output: torch.Tensor,
    query_5d: torch.Tensor,  # [batch_size, query_length, num_kv_heads, query_group_size, head_size]
    key_cache: torch.Tensor,
    value_cache: torch.Tensor,
    block_tables: torch.Tensor,
    context_lengths: torch.Tensor,
    softmax_scale: float,
    query_scale_5d: torch.Tensor,  # [num_seqs, query_length, num_kv_heads, query_group_size, 1](per-token) or [1](per-tensor) or None
    key_scale: torch.Tensor,
    value_scale: torch.Tensor,
    query_seq_len: int,
    query_group_size: int,
    head_size: int,
    num_sequences: int,
    num_kv_heads: int,
    max_context_partition_num: int,
    compute_type: torch.dtype,
    kv_block_size: int,
    context_partition_size: int,
    query_quant_mode: int,
    kv_quant_mode: int,
    fp8_max_value: float,
    value_transposed: int,
    is_causal: int,
    cdna_version: int,
    md_name: str,
    func_name: str = None,
):
    """
    Compile and run the compiled attention kernel with perftest timing
    """
    func = compile_attention_kernel(
        compute_type=compute_type,
        query_seq_len=query_seq_len,
        one_query_group_size=query_group_size,
        head_size=head_size,
        kv_block_size=kv_block_size,
        context_partition_size=context_partition_size,
        query_quant_mode=query_quant_mode,
        kv_quant_mode=kv_quant_mode,
        fp8_max_value=fp8_max_value,
        value_transposed=int(value_transposed),
        is_causal=int(is_causal),
        cdna_version=cdna_version,
        md_name=md_name,
        func_name=func_name,
    )

    # Configure query quantization
    stride_query_scale_bs = 0
    stride_query_scale_qlen = 0
    stride_query_scale_kv_head = 0
    if query_scale_5d is not None:
        assert (
            isinstance(query_scale_5d, torch.Tensor)
            and query_scale_5d.dtype == aiter.dtypes.fp32
        ), f"query_scale tensor only support dtype == {aiter.dtypes.fp32}, but got query_scale.dtype == {query_scale_5d.dtype}"

        if query_scale_5d.numel() == 1:
            pass
        else:
            # Per-token quantization (5D format)
            assert (
                len(query_scale_5d.shape) == 5
            ), f"Expected 5D query_scale tensor, but got shape {query_scale_5d.shape}"
            stride_query_scale_bs = query_scale_5d.stride(0)
            stride_query_scale_qlen = query_scale_5d.stride(1)
            stride_query_scale_kv_head = query_scale_5d.stride(2)

    # Configure KV quantization
    key_scale_stride_0 = 0
    key_scale_stride_1 = 0
    if key_scale is not None and value_scale is not None:
        assert (
            isinstance(key_scale, torch.Tensor) and key_scale.dtype == aiter.dtypes.fp32
        ), f"key_scale tensor only support dtype == {aiter.dtypes.fp32}, but got key_scale.dtype == {key_scale.dtype}"
        assert (
            isinstance(value_scale, torch.Tensor)
            and value_scale.dtype == aiter.dtypes.fp32
        ), f"value_scale tensor only support dtype == {aiter.dtypes.fp32}, but got value_scale.dtype == {value_scale.dtype}"

        if key_scale.numel() == 1:
            pass
        else:
            # Per-token quantization
            assert (
                len(key_scale.shape) == 4
            ), f"Expected 4D key_scale tensor, but got shape {key_scale.shape}"
            assert (
                key_scale.shape[-1] == 1
            ), f"Expected key_scale.shape[-1] == 1, but got key_scale.shape[-1]={key_scale.shape[-1]}"
            key_scale_stride_0 = key_scale.stride(0)
            key_scale_stride_1 = key_scale.stride(1)

        # Validate KV scale shape consistency
        assert (
            key_scale.shape == value_scale.shape
        ), f"Key and value scales must have same shape, but got key: {key_scale.shape}, value: {value_scale.shape}"

    func(
        *torch_to_c_types(
            exp_sums,
            max_logits,
            temporary_output,
            query_5d,
            key_cache,
            value_cache,
            block_tables,
            context_lengths,
            softmax_scale,
            query_scale_5d,
            key_scale,
            value_scale,
            exp_sums.stride(0),
            exp_sums.stride(1),
            exp_sums.stride(2),
            temporary_output.stride(0),
            temporary_output.stride(1),
            temporary_output.stride(2),
            temporary_output.stride(3),
            # 5D query strides
            query_5d.stride(0),
            query_5d.stride(1),
            query_5d.stride(2),
            query_5d.stride(3),
            key_cache.stride(0),
            key_cache.stride(1),
            key_cache.stride(2),
            key_cache.stride(3),
            value_cache.stride(0),
            value_cache.stride(1),
            value_cache.stride(2),
            block_tables.stride(0),
            # 5D query_scale strides
            stride_query_scale_bs,
            stride_query_scale_qlen,
            stride_query_scale_kv_head,
            key_scale_stride_0,
            key_scale_stride_1,
            head_size,
            num_sequences,
            num_kv_heads,
            max_context_partition_num,
            torch.cuda.current_stream(exp_sums.device),
        )
    )


@perftest()
def run_direct_attention_kernel(
    exp_sums: torch.Tensor,
    max_logits: torch.Tensor,
    temporary_output: torch.Tensor,
    query_5d: torch.Tensor,  # [batch_size, query_length, num_kv_heads, query_group_size, head_size]
    key_cache: torch.Tensor,
    value_cache: torch.Tensor,
    block_tables: torch.Tensor,
    context_lengths: torch.Tensor,
    softmax_scale: float,
    query_scale_5d: torch.Tensor,  # [num_seqs, query_length, num_kv_heads, query_group_size, 1](per-token) or [1](per-tensor) or None
    key_scale: torch.Tensor,
    value_scale: torch.Tensor,
    compute_type: torch.dtype,
    query_seq_len: int,
    query_group_size: int,
    head_size: int,
    kv_block_size: int,
    context_partition_size: int,
    query_quant_mode: int,
    kv_quant_mode: int,
    fp8_max_value: float,
    value_transposed: int,
    is_causal: int,
    cdna_version: int,
):
    """
    Directly call the attention kernel from pa_decode_gluon.py with perftest timing
    """
    # Convert compute_type from torch.dtype to tl.dtype
    compute_type_tl = TORCH_TO_TL_DTYPE[compute_type]

    num_seqs = exp_sums.shape[0]
    num_kv_heads = exp_sums.shape[1]
    max_context_partition_num = exp_sums.shape[2]
    # Configure grid
    grid = (num_seqs, num_kv_heads, max_context_partition_num)

    # Configure query quantization
    stride_query_scale_bs = 0
    stride_query_scale_qlen = 0
    stride_query_scale_kv_head = 0
    if query_scale_5d is not None:
        assert (
            isinstance(query_scale_5d, torch.Tensor)
            and query_scale_5d.dtype == aiter.dtypes.fp32
        ), f"query_scale tensor only support dtype == {aiter.dtypes.fp32}, but got query_scale.dtype == {query_scale_5d.dtype}"

        if query_scale_5d.numel() == 1:
            pass
        else:
            # Per-token quantization (5D format)
            assert (
                len(query_scale_5d.shape) == 5
            ), f"Expected 5D query_scale tensor, but got shape {query_scale_5d.shape}"
            stride_query_scale_bs = query_scale_5d.stride(0)
            stride_query_scale_qlen = query_scale_5d.stride(1)
            stride_query_scale_kv_head = query_scale_5d.stride(2)

    # Configure KV quantization
    key_scale_stride_0 = 0
    key_scale_stride_1 = 0
    if key_scale is not None and value_scale is not None:
        assert (
            isinstance(key_scale, torch.Tensor) and key_scale.dtype == aiter.dtypes.fp32
        ), f"key_scale tensor only support dtype == {aiter.dtypes.fp32}, but got key_scale.dtype == {key_scale.dtype}"
        assert (
            isinstance(value_scale, torch.Tensor)
            and value_scale.dtype == aiter.dtypes.fp32
        ), f"value_scale tensor only support dtype == {aiter.dtypes.fp32}, but got value_scale.dtype == {value_scale.dtype}"

        if key_scale.numel() == 1:
            pass
        else:
            # Per-token quantization
            assert (
                len(key_scale.shape) == 4
            ), f"Expected 4D key_scale tensor, but got shape {key_scale.shape}"
            assert (
                key_scale.shape[-1] == 1
            ), f"Expected key_scale.shape[-1] == 1, but got key_scale.shape[-1]={key_scale.shape[-1]}"
            key_scale_stride_0 = key_scale.stride(0)
            key_scale_stride_1 = key_scale.stride(1)

        # Validate KV scale shape consistency
        assert (
            key_scale.shape == value_scale.shape
        ), f"Key and value scales must have same shape, but got key: {key_scale.shape}, value: {value_scale.shape}"

    # Production path - select and launch appropriate kernel
    equi_query_group_size_pow2 = triton.next_power_of_2(
        query_seq_len
    ) * triton.next_power_of_2(query_group_size)
    # KV_COMPUTE_BLOCK_SIZE defaults to CONTEXT_PARTITION_SIZE (same as wrapper function)
    kv_compute_block_size = context_partition_size
    waves_per_eu = 1

    # Use standard kernel
    kernel = paged_attention_decode_v2_gluon_dot_kernel
    # Select kernel implementation based on block size
    if kv_block_size > context_partition_size:
        # Use large block kernel
        kernel = paged_attention_decode_v2_gluon_large_block_dot_kernel
        if value_transposed:
            # Use smaller compute block size for better performance with transposed values
            kv_compute_block_size = 128
    else:
        # Use standard kernel for normal block sizes
        # Configure waves per EU based on query group size
        if equi_query_group_size_pow2 == 64:
            waves_per_eu = 3
        else:
            waves_per_eu = 4

    # Launch the kernel directly (following _paged_attention_decode_v2_with_dot_kernel_reshape_wrapper)
    kernel[grid](
        exp_sums,
        max_logits,
        temporary_output,
        query_5d,
        key_cache,
        value_cache,
        block_tables,
        context_lengths,
        softmax_scale,
        query_scale_5d,
        key_scale,
        value_scale,
        exp_sums.stride(0),
        exp_sums.stride(1),
        exp_sums.stride(2),
        temporary_output.stride(0),
        temporary_output.stride(1),
        temporary_output.stride(2),
        temporary_output.stride(3),
        # 5D query strides
        query_5d.stride(0),
        query_5d.stride(1),
        query_5d.stride(2),
        query_5d.stride(3),
        key_cache.stride(0),
        key_cache.stride(1),
        key_cache.stride(2),
        key_cache.stride(3),
        value_cache.stride(0),
        value_cache.stride(1),
        value_cache.stride(2),
        block_tables.stride(0),
        stride_query_scale_bs,
        stride_query_scale_qlen,
        stride_query_scale_kv_head,
        key_scale_stride_0,
        key_scale_stride_1,
        head_size=head_size,
        num_seqs=num_seqs,
        num_kv_heads=num_kv_heads,
        max_context_partition_num=max_context_partition_num,
        COMPUTE_TYPE=compute_type_tl,
        QUERY_SEQ_LEN=query_seq_len,
        ONE_QUERY_GROUP_SIZE=query_group_size,
        HEAD_SIZE_POW2=triton.next_power_of_2(head_size),
        KV_BLOCK_SIZE=kv_block_size,
        CONTEXT_PARTITION_SIZE=context_partition_size,
        KV_COMPUTE_BLOCK_SIZE=kv_compute_block_size,
        QUERY_QUANT_MODE=query_quant_mode,
        KV_QUANT_MODE=kv_quant_mode,
        FP8_MAX_VALUE=fp8_max_value,
        VALUE_TRANSPOSED=value_transposed,
        IS_CAUSAL=is_causal,
        CDNA_VERSION=cdna_version,
        waves_per_eu=waves_per_eu,
        num_stages=1,
    )


def test_attention_kernel(kernel_type: str = "compiled"):
    """Test the attention kernel with provided parameters.

    Args:
        kernel_type: Type of kernel to test - "compiled" or "direct"
    """
    # Check CDNA version compatibility
    cdna_version = get_cdna_version()
    assert cdna_version in [
        3,
        4,
    ], f"test_attention_kernel only supports gfx942 (CDNA3) and gfx950 (CDNA4) now, but got {arch_info.get_arch()}"

    print(f"\n=== Testing Attention Kernel (Type: {kernel_type}) ===")
    setup_seed(123)
    device = "cuda:0"
    max_context_length = 16384
    compute_type = aiter.dtypes.bf16
    quant_q, quant_kv = [False, False]
    # compute_type = aiter.dtypes.fp8
    # quant_q, quant_kv = [True, True]
    data_type = compute_type
    # quant_mode = "per_token"
    quant_mode = "per_tensor"
    if compute_type == aiter.dtypes.fp8:
        data_type = aiter.dtypes.bf16

    context_length = 2048
    batch_size = 128
    # batch_size = 128 * 8
    num_heads = (8, 1)
    # num_heads = (64, 8)
    # num_heads = (16, 1)
    # num_heads = (64, 4)
    head_size = 128
    block_size = 16
    query_length = 1
    query_quant_mode = -1
    kv_quant_mode = -1
    context_partition_size = 256
    trans_v = False
    kv_varlen = False

    # ==================== FP8 CONFIGURATION ====================
    fp8_max_value = 1.0
    if quant_kv:
        fp8_max_value = torch.finfo(aiter.dtypes.fp8).max

    # Derived parameters
    num_query_heads, num_kv_heads = num_heads
    query_seq_len = query_length
    query_group_size = num_query_heads // num_kv_heads
    kv_block_size = block_size

    # Set quantization modes based on quant_mode parameter
    if quant_q:
        query_quant_mode = (
            1 if quant_mode == "per_token" else 0
        )  # 1 for per-token, 0 for per-tensor
    if quant_kv:
        kv_quant_mode = (
            1 if quant_mode == "per_token" else 0
        )  # 1 for per-token, 0 for per-tensor
    value_transposed = 1 if trans_v else 0

    # Determine if causal masking is needed
    is_causal = query_length > 1
    num_sequences = batch_size
    max_context_partition_num = (
        context_length + context_partition_size - 1
    ) // context_partition_size
    softmax_scale = 1.0 / (head_size**0.5)

    # Calculate block table dimensions
    max_blocks_per_sequence = (max_context_length + block_size - 1) // block_size
    total_blocks = max_blocks_per_sequence * batch_size
    blocks_per_sequence = (context_length + block_size - 1) // block_size

    # Create intermediate tensors for attention computation
    equivalent_query_group_size = query_length * (num_query_heads // num_kv_heads)
    intermediate_shape = (
        num_sequences,
        num_kv_heads,
        max_context_partition_num,
        equivalent_query_group_size,
    )

    exp_sums = torch.empty(intermediate_shape, dtype=torch.float32, device=device)
    max_logits = torch.empty(intermediate_shape, dtype=torch.float32, device=device)
    temporary_output = torch.empty(
        *intermediate_shape, head_size, dtype=data_type, device=device
    )

    num_query_heads, num_kv_heads = query_group_size, 1
    # Create query tensor following reference approach
    total_queries = batch_size * query_length
    qkv_tensor = torch.randn(
        total_queries,
        num_query_heads + 2 * num_kv_heads,
        head_size,
        dtype=data_type,
        device=device,
    )
    query, key, value = torch.split(
        qkv_tensor, [num_query_heads, num_kv_heads, num_kv_heads], dim=1
    )
    query.uniform_(*UNIFORM_RANGE)

    # Create key and value caches
    # itemsize should be 2 for bf16/fp16, 1 for fp8
    kv_itemsize = 1 if quant_kv else 2
    key_caches, value_caches = create_kv_cache(
        total_blocks,
        block_size,
        1,
        num_kv_heads,
        head_size,
        "auto",
        data_type,
        0,  # seed
        device,
        kv_itemsize,  # itemsize - critical for correct key_cache shape
    )
    key_cache, value_cache = key_caches[0], value_caches[0]
    num_query_heads, num_kv_heads = num_heads
    query = query.repeat(1, num_kv_heads, 1).contiguous()
    key_cache = key_cache.repeat(1, num_kv_heads, 1, 1, 1).contiguous()
    value_cache = value_cache.repeat(1, num_kv_heads, 1, 1).contiguous()

    # Quantization based on mode
    if quant_q:
        if quant_mode == "per_token":
            # Per-token quantization for query
            quantized_query, query_scale_factors = pertoken_quant(
                query, quant_dtype=aiter.dtypes.fp8
            )
        else:
            # Per-tensor quantization for query
            quantized_query, query_scale_factors = per_tensor_quant(
                query, quant_dtype=aiter.dtypes.fp8
            )
    else:
        quantized_query = query
        query_scale_factors = None

    if quant_kv:
        if quant_mode == "per_token":
            # Per-token quantization for KV cache
            (
                quantized_keys,
                key_scale_factors_flat,
                quantized_values,
                value_scale_factors_flat,
                key_scale_original,
                value_scale_original,
            ) = quantize_kv_cache_symmetric(
                key_cache, value_cache, quant_dtype=aiter.dtypes.fp8
            )
        else:
            # Per-tensor quantization for KV cache
            (
                quantized_keys,
                key_scale_factors_flat,
                quantized_values,
                value_scale_factors_flat,
                key_scale_original,
                value_scale_original,
            ) = quantize_kv_cache_per_tensor(
                key_cache, value_cache, quant_dtype=aiter.dtypes.fp8
            )
    else:
        quantized_keys = key_cache
        quantized_values = value_cache
        key_scale_original = None
        value_scale_original = None

    # Reshape query to 5D format for kernel
    # [batch_size * query_length, num_query_heads, head_size] -> [batch_size, query_length, num_kv_heads, query_group_size, head_size]
    quantized_query_5d = quantized_query.reshape(
        batch_size, query_length, num_kv_heads, query_group_size, head_size
    ).contiguous()

    # Reshape query_scale to 5D format if per-token quantization
    query_scale_5d = query_scale_factors
    if query_scale_factors is not None and len(query_scale_factors.shape) > 1:
        # per-token quantization: [batch_size * query_length, num_query_heads, 1] -> [batch_size, query_length, num_kv_heads, query_group_size, 1]
        query_scale_5d = query_scale_factors.reshape(
            batch_size, query_length, num_kv_heads, query_group_size, 1
        ).contiguous()

    # Transpose value cache if required
    if trans_v:
        quantized_values = shuffle_value_cache_layout(quantized_values)

    # Create sequence lengths and block tables
    if kv_varlen:
        kv_len_list = [
            random.randint(query_length, context_length) for _ in range(batch_size)
        ]
    else:
        kv_len_list = [context_length] * batch_size
    context_lengths = torch.tensor(kv_len_list, dtype=torch.int32, device=device)

    block_tables_list = []
    for _ in range(batch_size):
        block_table = [
            random.randint(0, total_blocks - 1) for _ in range(blocks_per_sequence)
        ]
        block_tables_list.append(block_table)
    block_tables = torch.tensor(block_tables_list, dtype=torch.int32, device=device)

    # Assign to variables used by the kernel
    query_5d = quantized_query_5d
    key_cache = quantized_keys
    value_cache = quantized_values
    query_scale = query_scale_5d
    key_scale = key_scale_original
    value_scale = value_scale_original

    # Execute kernel based on selected type
    if kernel_type == "compiled":
        # Compile and run the compiled kernel
        print("\n=== Running Compiled Kernel ===")
        _, compiled_time = run_compiled_attention_kernel(
            exp_sums,
            max_logits,
            temporary_output,
            query_5d,
            key_cache,
            value_cache,
            block_tables,
            context_lengths,
            softmax_scale,
            query_scale,
            key_scale,
            value_scale,
            query_seq_len,
            query_group_size,
            head_size,
            num_sequences,
            num_kv_heads,
            max_context_partition_num,
            compute_type=compute_type,
            kv_block_size=kv_block_size,
            context_partition_size=context_partition_size,
            query_quant_mode=query_quant_mode,
            kv_quant_mode=kv_quant_mode,
            fp8_max_value=fp8_max_value,
            value_transposed=value_transposed,
            is_causal=is_causal,
            cdna_version=cdna_version,
            md_name="pa_decode_attention_kernel",
        )
        print(f"Compiled kernel execution time: {compiled_time:.2f} us/iter")
    elif kernel_type == "direct":
        # Directly call the kernel from pa_decode_gluon.py
        print("\n=== Running Direct Kernel ===")
        _, direct_time = run_direct_attention_kernel(
            exp_sums,
            max_logits,
            temporary_output,
            query_5d,
            key_cache,
            value_cache,
            block_tables,
            context_lengths,
            softmax_scale,
            query_scale,
            key_scale,
            value_scale,
            compute_type,
            query_seq_len,
            query_group_size,
            head_size,
            kv_block_size,
            context_partition_size,
            query_quant_mode,
            kv_quant_mode,
            fp8_max_value,
            value_transposed,
            is_causal,
            cdna_version,
        )
        print(f"Direct kernel execution time: {direct_time:.2f} us/iter")
    else:
        raise ValueError(f"Unknown kernel type: {kernel_type}")

    # Check for NaN values
    exp_sums_nan_cnt = torch.isnan(exp_sums).sum().item()
    max_logits_nan_cnt = torch.isnan(max_logits).sum().item()
    temporary_output_nan_cnt = torch.isnan(temporary_output).sum().item()

    print("\n=== Checking Results ===")
    print(f"exp_sums NaN count: {exp_sums_nan_cnt}")
    print(f"max_logits NaN count: {max_logits_nan_cnt}")
    print(f"temporary_output NaN count: {temporary_output_nan_cnt}")

    # Compare with torch_attention_compute reference
    print("\n=== Comparing with Reference Implementation ===")

    # Convert 5D query to format expected by torch_attention_compute
    # query_5d: [batch_size, query_length, num_kv_heads, query_group_size, head_size]
    # torch_attention_compute expects: [num_seqs, num_q_heads, head_size] (actually 3D with combined batch*query_len)
    # We need to reshape to gluon format for the reference
    query_for_ref = query_5d.transpose(1, 2).reshape(
        batch_size, num_kv_heads * query_length * query_group_size, head_size
    )
    query_scale_for_ref = query_scale
    if query_scale is not None and len(query_scale.shape) == 5:
        query_scale_for_ref = query_scale.transpose(1, 2).reshape(
            batch_size, num_kv_heads * query_length * query_group_size, 1
        )

    # Run reference implementation
    ref_exp_sums, ref_max_logits, ref_temporary_output = torch_attention_compute(
        query=query_for_ref,
        key_cache=key_cache,
        value_cache=value_cache,
        block_tables=block_tables,
        context_lengths=context_lengths,
        softmax_scale=softmax_scale,
        q_seq_len=query_seq_len,
        query_scale=query_scale_for_ref,
        key_scale=key_scale,
        value_scale=value_scale,
        output_dtype=torch.bfloat16,
        kv_block_size=kv_block_size,
        context_partition_size=context_partition_size,
        is_causal=bool(is_causal),
    )

    # Compare exp_sums
    print("\n--- Comparing exp_sums ---")
    diff_exp_sums = (exp_sums - ref_exp_sums).abs()
    max_diff_exp_sums = diff_exp_sums.max().item()
    mean_diff_exp_sums = diff_exp_sums.mean().item()
    print(f"Max difference: {max_diff_exp_sums:.6e}")
    print(f"Mean difference: {mean_diff_exp_sums:.6e}")

    # Compare max_logits
    print("\n--- Comparing max_logits ---")
    diff_max_logits = (max_logits - ref_max_logits).abs()
    max_diff_max_logits = diff_max_logits.max().item()
    mean_diff_max_logits = diff_max_logits.mean().item()
    print(f"Max difference: {max_diff_max_logits:.6e}")
    print(f"Mean difference: {mean_diff_max_logits:.6e}")

    # Compare temporary_output
    print("\n--- Comparing temporary_output ---")
    diff_temporary_output = (temporary_output - ref_temporary_output).abs()
    max_diff_temporary_output = diff_temporary_output.max().item()
    mean_diff_temporary_output = diff_temporary_output.mean().item()
    print(f"Max difference: {max_diff_temporary_output:.6e}")
    print(f"Mean difference: {mean_diff_temporary_output:.6e}")

    # Detailed error analysis for temporary_output (largest tensor)
    if max_diff_temporary_output > 1e-4:
        print("\n=== Detailed Error Analysis (temporary_output) ===")
        # Find top 5 differences
        flat_diff = diff_temporary_output.flatten()
        top_k = 5
        top_k_indices = torch.topk(flat_diff, top_k).indices

        print(f"Top {top_k} differences:")
        for i in range(top_k):
            idx = top_k_indices[i]
            orig_idx = np.unravel_index(idx.cpu().numpy(), temporary_output.shape)
            print(
                f"  Position {orig_idx}: kernel={temporary_output[orig_idx].item():.6f}, ref={ref_temporary_output[orig_idx].item():.6f}, diff={flat_diff[idx].item():.6e}"
            )

    # Test results
    tolerance = 5e-3
    all_passed = True

    print("\n=== Test Results ===")
    if max_diff_exp_sums < tolerance:
        print(
            f"✅ exp_sums TEST PASSED: Max difference ({max_diff_exp_sums:.6e}) < tolerance ({tolerance})"
        )
    else:
        print(
            f"❌ exp_sums TEST FAILED: Max difference ({max_diff_exp_sums:.6e}) >= tolerance ({tolerance})"
        )
        all_passed = False

    if max_diff_max_logits < tolerance:
        print(
            f"✅ max_logits TEST PASSED: Max difference ({max_diff_max_logits:.6e}) < tolerance ({tolerance})"
        )
    else:
        print(
            f"❌ max_logits TEST FAILED: Max difference ({max_diff_max_logits:.6e}) >= tolerance ({tolerance})"
        )
        all_passed = False

    if max_diff_temporary_output < tolerance:
        print(
            f"✅ temporary_output TEST PASSED: Max difference ({max_diff_temporary_output:.6e}) < tolerance ({tolerance})"
        )
    else:
        print(
            f"❌ temporary_output TEST FAILED: Max difference ({max_diff_temporary_output:.6e}) >= tolerance ({tolerance})"
        )
        all_passed = False

    # MD5 hashes for verification
    ref_exp_sums_md5 = tensor_to_hash(ref_exp_sums)
    ref_max_logits_md5 = tensor_to_hash(ref_max_logits)
    ref_temporary_output_md5 = tensor_to_hash(ref_temporary_output)
    exp_sums_md5 = tensor_to_hash(exp_sums)
    # exp_sums_sha256 = tensor_to_hash(exp_sums, 'sha256')
    max_logits_md5 = tensor_to_hash(max_logits)
    temporary_output_md5 = tensor_to_hash(temporary_output)
    print(f"Reference exp_sums MD5: {ref_exp_sums_md5}")
    print(f"Reference max_logits MD5: {ref_max_logits_md5}")
    print(f"Reference temporary_output MD5: {ref_temporary_output_md5}")
    print(f"exp_sums MD5: {exp_sums_md5}")
    # print(f"exp_sums SHA256: {exp_sums_sha256}")
    print(f"max_logits MD5: {max_logits_md5}")
    print(f"temporary_output MD5: {temporary_output_md5}")

    # Test result
    if all_passed:
        print(
            "\n✅ OVERALL TEST PASSED: All comparisons within tolerance and no NaN values detected"
        )
        return True
    else:
        print(
            "\n❌ OVERALL TEST FAILED: Some comparisons failed or NaN values detected"
        )
        return False


def main():
    """Main function with command line argument parsing."""
    parser = argparse.ArgumentParser(description="Test paged attention kernel")
    parser.add_argument(
        "--kernel-type",
        type=str,
        choices=["compiled", "direct"],
        default="compiled",
        help="Type of kernel to test: 'compiled' (default) or 'direct'",
    )
    parser.add_argument(
        "--num-iters",
        type=int,
        default=101,
        help="Number of iterations for performance testing",
    )
    parser.add_argument(
        "--num-warmup", type=int, default=2, help="Number of warmup iterations"
    )

    args = parser.parse_args()

    # Run the test
    result = test_attention_kernel(kernel_type=args.kernel_type)
    sys.exit(0 if result else 1)


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: MIT
# Copyright (C) 2024-2025, Advanced Micro Devices, Inc. All rights reserved.
import functools
import os
import re
import subprocess

from cpp_extension import executable_path
from torch_guard import torch_compile_guard

GFX_MAP = {
    0: "native",
    1: "gfx90a",
    2: "gfx908",
    3: "gfx940",
    4: "gfx941",
    5: "gfx942",
    6: "gfx945",
    7: "gfx1100",
    8: "gfx950",
    9: "gfx1101",
    10: "gfx1102",
    11: "gfx1103",
    12: "gfx1150",
    13: "gfx1151",
    14: "gfx1152",
    15: "gfx1153",
    16: "gfx1200",
    17: "gfx1201",
}


@functools.lru_cache(maxsize=1)
def _detect_native() -> list[str]:
    try:
        rocminfo = executable_path("rocminfo")
        result = subprocess.run(
            [rocminfo],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            check=True,
        )
        for line in result.stdout.splitlines():
            if "gfx" in line.lower():
                return [line.split(":", 1)[-1].strip()]
    except Exception as e:
        raise RuntimeError(f"Get GPU arch from rocminfo failed: {e}") from e
    raise RuntimeError("No gfx arch found in rocminfo output.")


@torch_compile_guard()
def get_gfx_custom_op() -> int:
    return get_gfx_custom_op_core()


@functools.lru_cache(maxsize=10)
def get_gfx_custom_op_core() -> int:
    gfx = os.getenv("GPU_ARCHS", "native")
    gfx_mapping = {v: k for k, v in GFX_MAP.items()}
    # gfx = os.getenv("GPU_ARCHS", "native")
    if gfx == "native":
        try:
            rocminfo = executable_path("rocminfo")
            result = subprocess.run(
                [rocminfo], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
            )
            output = result.stdout
            for line in output.split("\n"):
                if "gfx" in line.lower():
                    try:
                        return gfx_mapping[line.split(":")[-1].strip()]
                    except KeyError:
                        raise KeyError(
                            f'Unknown GPU architecture: {line.split(":")[-1].strip()}. '
                            f"Supported architectures: {list(gfx_mapping.keys())}"
                        )

        except Exception as e:
            raise RuntimeError(f"Get GPU arch from rocminfo failed {str(e)}")
    elif ";" in gfx:
        gfx = gfx.split(";")[-1]
    try:
        return gfx_mapping[gfx]
    except KeyError:
        raise KeyError(
            f"Unknown GPU architecture: {gfx}. "
            f"Supported architectures: {list(gfx_mapping.keys())}"
        )


@functools.lru_cache(maxsize=1)
def get_gfx():
    gfx_num = get_gfx_custom_op()
    return GFX_MAP.get(gfx_num, "unknown")


@functools.lru_cache(maxsize=1)
def get_gfx_list() -> list[str]:

    gfx_env = os.getenv("GPU_ARCHS", "native")
    if gfx_env == "native":
        try:
            gfxs = _detect_native()
        except RuntimeError:
            gfxs = ["cpu"]
    else:
        gfxs = [g.strip() for g in gfx_env.split(";") if g.strip()]
    os.environ["AITER_GPU_ARCHS"] = ";".join(gfxs)

    return gfxs


@torch_compile_guard()
def get_cu_num_custom_op() -> int:
    cu_num = int(os.getenv("CU_NUM", 0))
    if cu_num == 0:
        try:
            rocminfo = executable_path("rocminfo")
            result = subprocess.run(
                [rocminfo], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
            )
            output = result.stdout
            devices = re.split(r"Agent\s*\d+", output)
            gpu_compute_units = []
            for device in devices:
                for line in device.split("\n"):
                    if "Device Type" in line and line.find("GPU") != -1:
                        match = re.search(r"Compute Unit\s*:\s*(\d+)", device)
                        if match:
                            gpu_compute_units.append(int(match.group(1)))
                        break
        except Exception as e:
            raise RuntimeError(f"Get GPU Compute Unit from rocminfo failed {str(e)}")
        assert len(set(gpu_compute_units)) == 1
        cu_num = gpu_compute_units[0]
    return cu_num


@functools.lru_cache(maxsize=1)
def get_cu_num():
    cu_num = get_cu_num_custom_op()
    return cu_num


def get_device_name():
    gfx = get_gfx()

    if gfx == "gfx942":
        cu = get_cu_num()
        if cu == 304:
            return "MI300"
        elif cu == 80 or cu == 64:
            return "MI308"
    elif gfx == "gfx950":
        return "MI350"
    else:
        raise RuntimeError("Unsupported gfx")



# SPDX-License-Identifier: MIT
# Copyright (C) 2024-2026, Advanced Micro Devices, Inc. All rights reserved.

from typing import Optional
import os
import torch
import triton
from aiter.ops.triton.utils.logger import AiterTritonLogger
from aiter.ops.triton._triton_kernels.gemm.fused.fused_gemm_afp4wfp4_a16w16 import (
    _fused_gemm_afp4wfp4_a16w16_kernel,
    _fused_gemm_afp4wfp4_preshuffle_a16w16_kernel,
    _fused_gemm_afp4wfp4_a16w16_reduce_kernel,
    _get_config,
)
from aiter.ops.triton.gemm.basic.gemm_afp4wfp4 import get_splitk
from aiter.ops.triton.utils.core import AITER_TRITON_CONFIGS_PATH

from aiter.utility.triton.triton_metadata_redirect import AOTMetadataContext

_LOGGER = AiterTritonLogger()


def fused_gemm_afp4wfp4_a16w16(
    x_fp4: torch.Tensor,
    w_fp4: torch.Tensor,
    x_fp4_scale: torch.Tensor,
    w_fp4_scale: torch.Tensor,
    x_bf16: torch.Tensor,
    w_bf16: torch.Tensor,
    is_fp4_preshuffled: bool = True,
    bias_fp4: Optional[torch.Tensor] = None,
    bias_bf16: Optional[torch.Tensor] = None,
    dtype: Optional[float] = torch.bfloat16,
    y_fp4: Optional[torch.Tensor] = None,
    y_bf16: Optional[torch.Tensor] = None,
    skip_reduce: Optional[bool] = False,
    config: Optional[dict] = None,
    use_aot: Optional[bool] = True,
):
    """
    Computes the 8 bit matmul Y = X x WT + B using the block-scale quantization approach for x_fp4 and w_fp4.
    Computes the 16 bit matmul Y = X x WT + B for x_bf16 and w_bf16

    This fusion is primarily aiming for fusing the gate up-projections and MOE gating:
        gate up-projections: (M, K) x (2N, K) = (M, 2N)
        MOE gating: (M, K) x (N, K) + (N, ) = (M, N)

    Key parameters:
    - x_fp4: Matrix X with shape (M, K).
    - w_fp4: Matrix W with shape (N_fp4, K).
    - x_fp4_scale: Scale tensor for X with shape (M, K // 32)
    - w_fp4_scale: Scale tensor for W with shape (N, K // 32)
    - x_bf16: Matrix X with shape (M, K).
    - w_bf16: Matrix W with shape (N_bf16, K).

    Note: M, N, K must be identical for x_fp4 and x_bf16, but the N-dim fow w_fp4 and w_bf16 can be different

    Returns:
    - y_fp4: The output matrix with shape (M, N_fp4).
    - y_bf16: The output matrix with shape (M, N_bf16).

    """
    _LOGGER.info(
        f"FUSED_GEMM_A8W8_BLOCKSCALE_A16W16: x_fp4={tuple(x_fp4.shape)} w_fp4={tuple(w_fp4.shape)} x_fp4_scale={tuple(x_fp4_scale.shape)} w_fp4_scale={tuple(w_fp4_scale.shape)} x_bf16={tuple(x_bf16.shape)} w_bf16={tuple(w_bf16.shape)}"
    )

    M, K = x_fp4.shape
    N_fp4, K = w_fp4.shape
    if is_fp4_preshuffled:
        N_fp4 = N_fp4 * 16
        K = K // 16
    M, _ = x_bf16.shape
    N_bf16, _ = w_bf16.shape

    # Check constraints.
    assert (
        x_fp4.shape[0] == x_bf16.shape[0]
    ), "M-dim should be identical for x_fp4 and x_bf16"
    assert (
        x_fp4.shape[1] * 2 == x_bf16.shape[1]
    ), "K-dim should be identical for x_fp4 and x_bf16"
    if is_fp4_preshuffled:
        assert x_fp4.shape[1] == w_fp4.shape[1] // 16, "Incompatible dimensions!!!"
    else:
        assert x_fp4.shape[1] == w_fp4.shape[1], "Incompatible dimensions!!!"
    assert x_bf16.shape[1] == w_bf16.shape[1], "Incompatible dimensions!!!"

    # Transpose w and w_scale
    if not is_fp4_preshuffled:
        w_fp4 = w_fp4.T
    w_bf16 = w_bf16.T

    if config is None:
        config, _ = _get_config(M, N_fp4, N_bf16, K, is_fp4_preshuffled)

    if config["NUM_KSPLIT"] > 1:
        SPLITK_BLOCK_SIZE, BLOCK_SIZE_K, NUM_KSPLIT = get_splitk(
            K, config["BLOCK_SIZE_K"], config["NUM_KSPLIT"]
        )

        config["SPLITK_BLOCK_SIZE"] = SPLITK_BLOCK_SIZE
        config["BLOCK_SIZE_K"] = BLOCK_SIZE_K
        config["NUM_KSPLIT"] = NUM_KSPLIT
        config["NUM_KSPLIT"] = triton.cdiv(K, config["SPLITK_BLOCK_SIZE"] // 2)
    else:
        config["SPLITK_BLOCK_SIZE"] = 2 * K

    if config["BLOCK_SIZE_K"] >= 2 * K:
        config["BLOCK_SIZE_K"] = triton.next_power_of_2(2 * K)
        config["SPLITK_BLOCK_SIZE"] = 2 * K
        config["NUM_KSPLIT"] = 1

    if y_fp4 is None and (config["NUM_KSPLIT"] == 1 or not skip_reduce):
        y_fp4 = torch.empty((M, N_fp4), dtype=dtype, device=x_fp4.device)

    if y_bf16 is None and (config["NUM_KSPLIT"] == 1 or not skip_reduce):
        y_bf16 = torch.empty((M, N_bf16), dtype=dtype, device=x_bf16.device)

    if config["NUM_KSPLIT"] > 1:
        y_fp4_pp = torch.empty(
            (config["NUM_KSPLIT"], M, N_fp4),
            dtype=torch.float32,
            device=x_fp4.device,
        )
        y_bf16_pp = torch.empty(
            (config["NUM_KSPLIT"], M, N_bf16),
            dtype=torch.float32,
            device=x_bf16.device,
        )
    else:
        y_fp4_pp = None
        y_bf16_pp = None

    config["BLOCK_SIZE_N"] = max(config["BLOCK_SIZE_N"], 32)
    if is_fp4_preshuffled:
        if M < 32:
            assert (
                config["BLOCK_SIZE_M"] <= 16
            ), "for M < 32, BLOCK_SIZE_M must be 16 or less as x_scale are assumed to be un-shuffled"
        else:
            assert (
                config["BLOCK_SIZE_M"] >= 32
            ), "for M >= 32, BLOCK_SIZE_M must be 32 or more as x_scale are assumed to be preshuffled"

    grid = lambda META: (  # noqa: E731
        (
            META["NUM_KSPLIT"]
            * triton.cdiv(M, META["BLOCK_SIZE_M"])
            * (
                triton.cdiv(N_fp4, META["BLOCK_SIZE_N"])
                + triton.cdiv(N_bf16, META["BLOCK_SIZE_N"])
            )
        ),
    )
    selected_kernel = (
        _fused_gemm_afp4wfp4_preshuffle_a16w16_kernel
        if is_fp4_preshuffled
        else _fused_gemm_afp4wfp4_a16w16_kernel
    )

    def selected_kernel_wrapper():
        selected_kernel[grid](
            x_fp4,
            w_fp4,
            bias_fp4,
            x_fp4_scale,
            w_fp4_scale,
            y_fp4 if config["NUM_KSPLIT"] == 1 else y_fp4_pp,
            x_bf16,
            w_bf16,
            bias_bf16,
            y_bf16 if config["NUM_KSPLIT"] == 1 else y_bf16_pp,
            M,
            N_fp4,
            N_bf16,
            K,
            x_fp4.stride(0),
            x_fp4.stride(1),
            w_fp4.stride(0),
            w_fp4.stride(1),
            x_fp4_scale.stride(0),
            x_fp4_scale.stride(1),
            w_fp4_scale.stride(0),
            w_fp4_scale.stride(1),
            0 if config["NUM_KSPLIT"] == 1 else y_fp4_pp.stride(0),
            y_fp4.stride(0) if config["NUM_KSPLIT"] == 1 else y_fp4_pp.stride(1),
            y_fp4.stride(1) if config["NUM_KSPLIT"] == 1 else y_fp4_pp.stride(2),
            x_bf16.stride(0),
            x_bf16.stride(1),
            w_bf16.stride(0),
            w_bf16.stride(1),
            0 if config["NUM_KSPLIT"] == 1 else y_bf16_pp.stride(0),
            y_bf16.stride(0) if config["NUM_KSPLIT"] == 1 else y_bf16_pp.stride(1),
            y_bf16.stride(1) if config["NUM_KSPLIT"] == 1 else y_bf16_pp.stride(2),
            ADD_BIAS_FP4=(bias_fp4 is not None),
            ADD_BIAS_BF16=(bias_bf16 is not None),
            SKIP_REDUCE=skip_reduce,
            **config,
        )

    M_POW2 = triton.next_power_of_2(M)
    if M < 32 and M_POW2 > 16:
        M_POW2 = 16
    metadata_pth = f"{AITER_TRITON_CONFIGS_PATH}/gemm/aot/{selected_kernel.fn.__name__}_M={M_POW2}-N4={N_fp4}-N16={N_bf16}-K={K*2}"
    if use_aot and os.path.exists(metadata_pth):
        with AOTMetadataContext(
            selected_kernel.fn.__name__,
            f"{metadata_pth}",
        ):
            selected_kernel_wrapper()
    else:
        selected_kernel_wrapper()

    if config["NUM_KSPLIT"] > 1:
        if skip_reduce:
            return y_fp4_pp, y_bf16_pp
        REDUCE_BLOCK_SIZE_M = 32
        REDUCE_BLOCK_SIZE_N = 32
        ACTUAL_KSPLIT = triton.cdiv(K, config["SPLITK_BLOCK_SIZE"] // 2)

        grid_reduce = (
            triton.cdiv(M, REDUCE_BLOCK_SIZE_M),
            triton.cdiv(N_fp4, REDUCE_BLOCK_SIZE_N)
            + triton.cdiv(N_bf16, REDUCE_BLOCK_SIZE_N),
        )
        _fused_gemm_afp4wfp4_a16w16_reduce_kernel[grid_reduce](
            bias_fp4,
            y_fp4_pp,
            y_fp4,
            bias_bf16,
            y_bf16_pp,
            y_bf16,
            M,
            N_fp4,
            N_bf16,
            y_fp4_pp.stride(0),
            y_fp4_pp.stride(1),
            y_fp4_pp.stride(2),
            y_fp4.stride(0),
            y_fp4.stride(1),
            y_bf16_pp.stride(0),
            y_bf16_pp.stride(1),
            y_bf16_pp.stride(2),
            y_bf16.stride(0),
            y_bf16.stride(1),
            REDUCE_BLOCK_SIZE_M,
            REDUCE_BLOCK_SIZE_N,
            ACTUAL_KSPLIT,
            triton.next_power_of_2(config["NUM_KSPLIT"]),
            ADD_BIAS_FP4=(bias_fp4 is not None),
            ADD_BIAS_BF16=(bias_bf16 is not None),
        )

    return y_fp4, y_bf16



import torch
import importlib.util
from .base_device_communicator import All2AllManagerBase, Cache
from functools import cache
from aiter import logger


@cache
def _has_module(module_name: str) -> bool:
    """Return True if *module_name* can be found in the current environment.
    The result is cached so that subsequent queries for the same module incur
    no additional overhead.
    """
    return importlib.util.find_spec(module_name) is not None


def has_mori() -> bool:
    """Whether the optional `mori` package is available."""
    return _has_module("mori")


class MoriAll2AllManager(All2AllManagerBase):
    def __init__(self, cpu_group):
        assert has_mori(), (
            "MoRI kernels not found. Please follow https://github.com/ROCm/mori/blob/main/README.md"
            " to install MoRI kernels."
        )  # noqa
        import mori

        super().__init__(cpu_group)
        self.handle_cache = Cache()

        torch._C._distributed_c10d._register_process_group("mori", cpu_group)
        mori.shmem.shmem_torch_process_group_init("mori")

    def _make_all2all_kwargs(
        self,
        rank: int,
        num_ep_ranks: int,
        input_dtype: torch.dtype,
        quant_dtype: torch.dtype,
        token_hidden_size: int,
        scale_dim: int,
        scale_type_size: int,
        max_num_tokens_per_dp_rank: int,
        num_local_experts: int,
        num_experts_per_token: int,
        gpu_per_node: int,
    ):
        import mori  # type: ignore[import-not-found]

        if not self.internode:
            # single node
            kernel_type = mori.ops.EpDispatchCombineKernelType.IntraNode
            warp_num_per_block = 16
            block_num = 80
            rdma_block_num = 0
        else:
            # multi node
            kernel_type = mori.ops.EpDispatchCombineKernelType.InterNodeV1
            warp_num_per_block = 16
            block_num = 32
            rdma_block_num = 16

        return dict(
            rank=rank,
            world_size=num_ep_ranks,
            data_type=quant_dtype,
            hidden_dim=token_hidden_size,
            scale_dim=scale_dim,
            scale_type_size=scale_type_size,
            max_token_type_size=input_dtype.itemsize,
            max_num_inp_token_per_rank=max_num_tokens_per_dp_rank,
            num_experts_per_rank=num_local_experts,
            num_experts_per_token=num_experts_per_token,
            warp_num_per_block=warp_num_per_block,
            block_num=block_num,
            kernel_type=kernel_type,
            rdma_block_num=rdma_block_num,
            gpu_per_node=gpu_per_node,
        )

    def _make_handle(self, **kwargs):
        import mori  # type: ignore[import-not-found]

        mori_config = mori.ops.EpDispatchCombineConfig(**kwargs)
        handle = mori.ops.EpDispatchCombineOp(mori_config)
        return handle

    def get_handle(self, kwargs):
        import mori  # type: ignore[import-not-found]

        mori_kwargs = self._make_all2all_kwargs(**kwargs)
        logger.debug("MoRI all2all args %s", mori_kwargs)
        handle: mori.ops.EpDispatchCombineOp = self.handle_cache.get_or_create(
            mori_kwargs, self._make_handle
        )
        return handle



import triton.language as tl
from aiter.ops.triton.utils._triton.kernel_repr import make_kernel_repr
from aiter.ops.triton.utils._triton.pid_preprocessing import pid_grid, remap_xcd
from aiter.ops.triton.utils.gemm_config_utils import (
    compute_splitk_params,
    get_gemm_config,
)

import triton

_gemm_a16w16_repr = make_kernel_repr(
    "_gemm_a16_w16_kernel",
    [
        "BLOCK_SIZE_M",
        "BLOCK_SIZE_N",
        "BLOCK_SIZE_K",
        "GROUP_SIZE_M",
        "NUM_KSPLIT",
        "SPLITK_BLOCK_SIZE",
        "EVEN_K",
        "GRID_MN",
        "cache_modifier",
        "activation",
        "use_activation",
        "ADD_BIAS",
        "SKIP_REDUCE",
    ],
)


_gemm_a16w16_reduce_repr = make_kernel_repr(
    "_gemm_a16w16_reduce_kernel",
    [
        "BLOCK_SIZE_M",
        "BLOCK_SIZE_N",
        "ACTUAL_KSPLIT",
        "MAX_KSPLIT",
        "activation",
        "use_activation",
        "ADD_BIAS",
    ],
)


@triton.heuristics(
    {
        "EVEN_K": lambda args: (args["K"] % (args["SPLITK_BLOCK_SIZE"]) == 0)
        and (args["SPLITK_BLOCK_SIZE"] % args["BLOCK_SIZE_K"] == 0),
        "GRID_MN": lambda args: triton.cdiv(args["M"], args["BLOCK_SIZE_M"])
        * triton.cdiv(args["N"], args["BLOCK_SIZE_N"]),
    }
)
@triton.jit(repr=_gemm_a16w16_repr)
def _gemm_a16_w16_kernel(
    a_ptr,
    b_ptr,
    bias_ptr,
    c_ptr,
    M,
    N,
    K,
    stride_am,
    stride_ak,
    stride_bk,
    stride_bn,
    stride_ck,
    stride_cm,
    stride_cn,
    # Meta-parameters
    BLOCK_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr,
    BLOCK_SIZE_K: tl.constexpr,
    GROUP_SIZE_M: tl.constexpr,
    NUM_KSPLIT: tl.constexpr,
    SPLITK_BLOCK_SIZE: tl.constexpr,
    EVEN_K: tl.constexpr,
    GRID_MN: tl.constexpr,
    cache_modifier: tl.constexpr,
    activation: tl.constexpr,
    use_activation: tl.constexpr,
    ADD_BIAS: tl.constexpr,
    SKIP_REDUCE: tl.constexpr,
):
    """Kernel for computing the matmul C = A x B.
    A has shape (M, K), B has shape (K, N) and C has shape (M, N)
    """

    tl.assume(stride_am > 0)
    tl.assume(stride_ak > 0)
    tl.assume(stride_bk > 0)
    tl.assume(stride_bn > 0)
    tl.assume(stride_ck > 0)
    tl.assume(stride_cm > 0)
    tl.assume(stride_cn > 0)

    # -----------------------------------------------------------
    # Map program ids `pid` to the block of C it should compute.
    # This is done in a grouped ordering to promote L2 data reuse.
    pid_unified = tl.program_id(axis=0)
    pid_unified = remap_xcd(pid_unified, GRID_MN * NUM_KSPLIT, NUM_XCDS=8)
    pid_k = pid_unified % NUM_KSPLIT
    pid = pid_unified // NUM_KSPLIT
    num_pid_m = tl.cdiv(M, BLOCK_SIZE_M)
    num_pid_n = tl.cdiv(N, BLOCK_SIZE_N)

    if NUM_KSPLIT == 1:
        pid_m, pid_n = pid_grid(pid, num_pid_m, num_pid_n, GROUP_SIZE_M=GROUP_SIZE_M)
    else:
        pid_m = pid // num_pid_n
        pid_n = pid % num_pid_n

    tl.assume(pid_m >= 0)
    tl.assume(pid_n >= 0)
    tl.assume(pid_k >= 0)

    split_k_start = pid_k * SPLITK_BLOCK_SIZE
    if split_k_start < K:
        # Create pointers for first block of A and B input matrices
        offs_k = tl.arange(0, BLOCK_SIZE_K)
        offs_k_split = split_k_start + offs_k
        offs_am = (pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)) % M
        offs_bn = (pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)) % N

        a_ptrs = a_ptr + (
            offs_am[:, None] * stride_am + offs_k_split[None, :] * stride_ak
        )
        b_ptrs = b_ptr + (
            offs_k_split[:, None] * stride_bk + offs_bn[None, :] * stride_bn
        )

        acc_dtype = tl.float32 if c_ptr.type.element_ty != tl.int8 else tl.int32
        if ADD_BIAS:
            if NUM_KSPLIT == 1 or (SKIP_REDUCE and pid_k == 0):
                accumulator = tl.load(bias_ptr + offs_bn).to(dtype=acc_dtype)
                accumulator = tl.broadcast_to(
                    accumulator[None, :], (BLOCK_SIZE_M, BLOCK_SIZE_N)
                )
            else:
                accumulator = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=acc_dtype)
        else:
            accumulator = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=acc_dtype)

        split_k_end = tl.minimum(split_k_start + SPLITK_BLOCK_SIZE, K)
        k_span = split_k_end - split_k_start
        num_k_iter = tl.cdiv(k_span, BLOCK_SIZE_K)

        for k in range(num_k_iter):
            # Load the next block of A and B, generate a mask by checking the K dimension.
            # If it is out of bounds, set it to 0.
            if EVEN_K:
                a = tl.load(a_ptrs)
                b = tl.load(b_ptrs, cache_modifier=cache_modifier)
            else:
                a = tl.load(
                    a_ptrs, mask=offs_k[None, :] < k_span - k * BLOCK_SIZE_K, other=0.0
                )
                b = tl.load(
                    b_ptrs,
                    mask=offs_k[:, None] < k_span - k * BLOCK_SIZE_K,
                    other=0.0,
                    cache_modifier=cache_modifier,
                )
            accumulator += tl.dot(a, b, input_precision="ieee")
            # Advance the ptrs to the next K block.
            a_ptrs += BLOCK_SIZE_K * stride_ak
            b_ptrs += BLOCK_SIZE_K * stride_bk

        if use_activation and NUM_KSPLIT == 1:
            accumulator = activation(accumulator)

        # Write back the block of the output matrix C with masks.
        c = accumulator.to(c_ptr.type.element_ty)
        offs_cm = pid_m.to(tl.int64) * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
        offs_cn = pid_n.to(tl.int64) * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
        c_ptrs = (
            c_ptr
            + stride_cm * offs_cm[:, None]
            + stride_cn * offs_cn[None, :]
            + pid_k * stride_ck
        )
        c_mask = (offs_cm[:, None] < M) & (offs_cn[None, :] < N)
        tl.store(c_ptrs, c, mask=c_mask)


@triton.jit(repr=_gemm_a16w16_reduce_repr)
def _gemm_a16w16_reduce_kernel(
    bias_ptr,
    c_in_ptr,
    c_out_ptr,
    M,
    N,
    stride_c_in_k,
    stride_c_in_m,
    stride_c_in_n,
    stride_c_out_m,
    stride_c_out_n,
    BLOCK_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr,
    ACTUAL_KSPLIT: tl.constexpr,
    MAX_KSPLIT: tl.constexpr,
    activation: tl.constexpr,
    use_activation: tl.constexpr,
    ADD_BIAS: tl.constexpr,
):

    tl.assume(stride_c_in_k > 0)
    tl.assume(stride_c_in_m > 0)
    tl.assume(stride_c_in_n > 0)
    tl.assume(stride_c_out_m > 0)
    tl.assume(stride_c_out_n > 0)

    pid_m = tl.program_id(axis=0)
    pid_n = tl.program_id(axis=1)

    tl.assume(pid_m > 0)
    tl.assume(pid_n > 0)

    offs_m = (pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)) % M
    offs_n = (pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)) % N
    offs_k = tl.arange(0, MAX_KSPLIT)
    c_in_ptrs = (
        c_in_ptr
        + (offs_k[:, None, None] * stride_c_in_k)
        + (offs_m[None, :, None] * stride_c_in_m)
        + (offs_n[None, None, :] * stride_c_in_n)
    )

    if ACTUAL_KSPLIT == MAX_KSPLIT:
        c = tl.load(c_in_ptrs)
    else:
        c = tl.load(c_in_ptrs, mask=offs_k[:, None, None] < ACTUAL_KSPLIT, other=0.0)
    c = tl.sum(c, axis=0)
    acc_dtype = tl.float32 if c_in_ptr.type.element_ty != tl.int8 else tl.int32
    if ADD_BIAS:
        bias = tl.load(bias_ptr + offs_n).to(dtype=acc_dtype)
        bias = tl.broadcast_to(bias[None, :], (BLOCK_SIZE_M, BLOCK_SIZE_N))
        c += bias

    if use_activation:
        c = activation(c)
    c = c.to(c_out_ptr.type.element_ty)

    c_out_ptrs = (
        c_out_ptr
        + (offs_m[:, None] * stride_c_out_m)
        + (offs_n[None, :] * stride_c_out_n)
    )

    tl.store(c_out_ptrs, c)


def _get_config(
    M: int,
    N: int,
    K: int,
):
    config, is_tunned = get_gemm_config("GEMM-A16W16", M, N, K)
    return compute_splitk_params(config, K), is_tunned



# SPDX-License-Identifier: MIT
# Copyright (C) 2024-2026, Advanced Micro Devices, Inc. All rights reserved.

from typing import Optional
import torch
import triton
from aiter.ops.triton.utils.logger import AiterTritonLogger
from aiter.ops.triton._triton_kernels.moe.moe_routing_sigmoid_top1_fused import (
    _routing_sigmoid_top1_kernel,
    _get_config,
)

_LOGGER = AiterTritonLogger()


def routing_sigmoid_top1(
    x, w, topk, fused_shared_experts=False, config: Optional[dict[str, any]] = None
):
    """
    Computes top-1 MoE routing with sigmoid activation for expert selection.

    Args:
        x (torch.Tensor): Input activations with shape (batch_size, seq_len, hidden_dim) or (M, K).
        w (torch.Tensor): Routing weights with shape (hidden_dim, num_experts).
        topk (int): Number of experts to select. Must be 1.
        fused_shared_experts (bool): Include shared expert (always selected) alongside top-1.
        config (Optional[dict]): Kernel tuning parameters (BLOCK_M, BLOCK_K).

    Returns:
        tuple: (topk_ids, topk_weights)
            - topk_ids (torch.Tensor): Selected expert IDs with shape (M, topk) or (M, topk+1) if fused_shared_experts.
            - topk_weights (torch.Tensor): Routing weights (sigmoid scores) with shape (M, topk) or (M, topk+1).
    """
    _LOGGER.info(
        f"ROUTING_SIGMOID_TOP1:  x={tuple(x.shape)}  w={tuple(w.shape)}  topk={topk} "
    )
    x = x.view(-1, x.shape[-1])

    assert topk == 1

    # M: batch_size x seq_len, K: hidden_dim, N: num_experts
    M, K = x.shape
    Kb, N = w.shape
    assert K == Kb

    _topk = topk
    if fused_shared_experts:
        _topk += 1

    # Output tensor
    topk_ids = torch.empty((M, _topk), device=x.device, dtype=torch.int32)
    topk_weights = torch.empty((M, _topk), device=x.device, dtype=torch.float32)

    config = _get_config(M, N, K)

    # Grid size
    def grid(META):
        return (triton.cdiv(M, META["BLOCK_M"]),)

    _routing_sigmoid_top1_kernel[grid](
        x,
        w,
        topk_ids,
        topk_weights,
        M,
        N,
        K,
        x.stride(0),
        x.stride(1),
        w.stride(0),
        w.stride(1),
        topk_ids.stride(0),
        topk_ids.stride(1),
        topk_weights.stride(0),
        topk_weights.stride(1),
        BLOCK_N=N,  # Set BLOCK_N to N
        TOPK=topk,
        FUSED_SHARED_EXPERTS=fused_shared_experts,
        **config,
    )

    return topk_ids, topk_weights



# SPDX-License-Identifier: MIT
# Copyright (C) 2025, Advanced Micro Devices, Inc. All rights reserved.

"""
Fused Communication + Computation Kernels

This submodule contains Triton kernels that fuse communication operations
with computation operations for improved performance.

Examples:
- reduce_scatter + rmsnorm + quant + all_gather
- all_reduce + rmsnorm + quant
- reduce_scatter + gemm + all_gather
"""

from .reduce_scatter_rmsnorm_quant_all_gather import (
    reduce_scatter_rmsnorm_quant_all_gather,
)

__all__ = [
    "reduce_scatter_rmsnorm_quant_all_gather",
]



import torch
from aiter.ops.triton._triton_kernels.fusions.fused_qk_concat import (
    _qk_rope_cat_kernel,
    _qk_cat_kernel,
)
from aiter.ops.triton.utils.logger import AiterTritonLogger

_LOGGER = AiterTritonLogger()


def fused_qk_cat(
    q1: torch.Tensor,
    q2: torch.Tensor,
    k1: torch.Tensor,
    k2: torch.Tensor,
):
    """
    Concat q1 with q2 and k1 with k2 along the last dimension

    Key parameters:
    - q1: Matrix X with shape (B, QH, D1).
    - q2: Matrix W with shape (B, QH, D2).
    - k1: Matrix X with shape (B, KH, D1).
    - k2: Matrix W with shape (B, KH, D2).

    QH must be multiple of KH

    Returns:
    - q_out: The output matrix with shape (B, QH, D1+D2).
    - k_out: The output matrix with shape (B, KH, D1+D2).
    """
    _LOGGER.info(
        f"FUSED_QK_CAT: q1={tuple(q1.shape)} q2={tuple(q2.shape)} k1={tuple(k1.shape)} k2={tuple(k2.shape)} "
    )
    b, qh, d1 = q1.shape
    b2, qh2, d2 = q2.shape
    bk, kh, dk1 = k1.shape
    bk2, kh2, dk2 = k2.shape
    assert (
        b == b2 == bk == bk2
    ), "q1 batch dimension should be identical across all inputs"
    assert qh == qh2, "Q head should be identical"
    assert kh == kh2, "K head should be identical"
    assert d1 == dk1, "D dimension of q1 and k1 should be identical"
    assert d2 == dk2, "D dimension of q2 and k2 should be identical"
    assert qh % kh == 0, "Number of Q heads must be multiple of number H heads"

    q_out = torch.empty((b, qh, d1 + d2), dtype=q1.dtype, device=q1.device)
    k_out = torch.empty((b, kh, d1 + d2), dtype=q1.dtype, device=q1.device)

    grid = (b, qh, 1)

    _qk_cat_kernel[grid](
        q1,
        q2,
        k1,
        k2,
        q_out,
        k_out,
        *q1.stride(),
        *q2.stride(),
        *k1.stride(),
        *k2.stride(),
        *q_out.stride(),
        *k_out.stride(),
        QH_PER_KH=qh // kh,
        BLOCK_D1=d1,
        BLOCK_D2=d2,
    )

    return q_out, k_out


def fused_qk_rope_cat(
    q_nope: torch.Tensor,
    q_pe: torch.Tensor,
    k_nope: torch.Tensor,
    k_pe: torch.Tensor,
    pos: torch.Tensor,
    cos: torch.Tensor,
    sin: torch.Tensor,
    is_neox: bool,
):
    """
    Perform RoPE on q_pe and k_pe and concat q_nope with q_pe and k_nope with k_pe along the last dimension

    Key parameters:
    - q_nope: Matrix X with shape (B, QH, D1).
    - q_pe: Matrix W with shape (B, QH, D2).
    - k_nope: Matrix X with shape (B, KH, D1).
    - k_pe: Matrix W with shape (B, KH, D2).

    QH must be multiple of KH

    Returns:
    - q_out: The output matrix with shape (B, QH, D1+D2).
    - k_out: The output matrix with shape (B, KH, D1+D2).
    """
    _LOGGER.info(
        f"FUSED_QK_ROPE_CAT: q_nope={tuple(q_nope.shape)} q_pe={tuple(q_pe.shape)} k_nope={tuple(k_nope.shape)} k_pe={tuple(k_pe.shape)} "
        + f"pos={tuple(pos.shape)} cos={tuple(cos.shape)} sin={tuple(sin.shape)}"
    )
    b, qh, d_nope = q_nope.shape
    b2, qh2, d_pe = q_pe.shape
    bk, kh, dk1 = k_nope.shape
    bk2, kh2, dk2 = k_pe.shape

    assert (
        b == b2 == bk == bk2
    ), "q1 batch dimension should be identical across all inputs"
    assert qh == qh2, "Q head should be identical"
    assert kh == kh2, "K head should be identical"
    assert d_nope == dk1, "D dimension of q_nope and k_nope should be identical"
    assert d_pe == dk2, "D dimension of q_pe and k_pe should be identical"
    assert qh % kh == 0, "Q heads must be multiple of H heads"
    d_freq = cos.shape[-1]
    assert (d_freq == d_pe // 2) or (
        d_freq == d_pe
    ), "cos/sin last dim should be the same or half of the qk last dim"
    reuse_freqs_front_part = d_freq == d_pe // 2

    q_out = torch.empty(
        (b, qh, d_nope + d_pe), dtype=q_nope.dtype, device=q_nope.device
    )
    k_out = torch.empty(
        (b, kh, d_nope + d_pe), dtype=q_nope.dtype, device=q_nope.device
    )

    grid = (b, qh, 1)

    _qk_rope_cat_kernel[grid](
        q_nope,
        q_pe,
        k_nope,
        k_pe,
        pos,
        cos,
        sin,
        q_out,
        k_out,
        *q_nope.stride(),
        *q_pe.stride(),
        *k_nope.stride(),
        *k_pe.stride(),
        pos.stride(0),
        cos.stride(0),
        cos.stride(-1),
        *q_out.stride(),
        *k_out.stride(),
        QH_PER_KH=qh // kh,
        REUSE_FREQS_FRONT_PART=reuse_freqs_front_part,
        IS_NEOX=is_neox,
        BLOCK_D_nope=d_nope,
        BLOCK_D_pe=d_pe,
        BLOCK_D_HALF_pe=d_pe // 2,
    )

    return q_out, k_out



# SPDX-License-Identifier: MIT
# Copyright (C) 2024-2026, Advanced Micro Devices, Inc. All rights reserved.
import torch
from ..jit.utils.chip_info import get_gfx
from ..ops.enum import QuantType
import argparse

defaultDtypes = {
    "gfx942": {"fp8": torch.float8_e4m3fnuz},
    "gfx950": {"fp8": torch.float8_e4m3fn},
    "gfx1250": {"fp8": torch.float8_e4m3fn},
}

_8bit_fallback = torch.uint8


def get_dtype_fp8():
    return defaultDtypes.get(get_gfx(), {"fp8": _8bit_fallback})["fp8"]


i4x2 = getattr(torch, "int4", _8bit_fallback)
fp4x2 = getattr(torch, "float4_e2m1fn_x2", _8bit_fallback)
fp8 = get_dtype_fp8()
fp8_e8m0 = getattr(torch, "float8_e8m0fnu", _8bit_fallback)
fp16 = torch.float16
bf16 = torch.bfloat16
fp32 = torch.float32
u32 = torch.uint32
i32 = torch.int32
i16 = torch.int16
i8 = torch.int8

d_dtypes = {
    "fp8": fp8,
    "fp8_e8m0": fp8_e8m0,
    "fp16": fp16,
    "bf16": bf16,
    "fp32": fp32,
    "i4x2": i4x2,
    "fp4x2": fp4x2,
    "u32": u32,
    "i32": i32,
    "i16": i16,
    "i8": i8,
}


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def str2tuple(v):
    """
    Convert string to int or tuple of ints.
    - "512" -> 512 (single value without comma returns int)
    - "512," -> (512,) (trailing comma returns tuple)
    - "512,1024" -> (512, 1024) (multiple values return tuple)
    """
    try:
        parts = [int(p.strip()) for p in v.strip("()").split(",") if p.strip()]
        # Return single value if only one element and no comma; otherwise return tuple
        if "," not in v and len(parts) == 1:
            return parts[0]
        return tuple(parts)
    except Exception as e:
        raise argparse.ArgumentTypeError(f"invalid format of input: {v}") from e


def str2Dtype(v):
    def _convert(s):
        if s.lower() == "none":
            return None
        elif s in d_dtypes:
            return d_dtypes[s]
        else:
            # Case-insensitive lookup for QuantType
            s_lower = s.lower()
            for name in dir(QuantType):
                if not name.startswith("_") and name.lower() == s_lower:
                    return getattr(QuantType, name)
            raise ValueError(f"'{s}' not in d_dtypes or QuantType")

    try:
        parts = [p.strip() for p in v.strip("()").split(",") if p.strip()]
        # Return single value if only one element and no comma; otherwise return tuple
        if len(parts) == 1 and "," not in v:
            return _convert(parts[0])
        return tuple(_convert(p) for p in parts)
    except Exception as e:
        raise argparse.ArgumentTypeError(f"invalid format of type: {v}") from e



from jinja2 import Template
from csrc.cpp_itfs.utils import (
    compile_template_op,
    transfer_hsaco,
    AITER_CORE_DIR,
    get_default_func_name,
    not_built,
    run_lib,
    CK_DIR,
)
import ctypes

MD_NAME = "asm_moe"
with open(f"{AITER_CORE_DIR}/csrc/cpp_itfs/moe/asm_moe.cpp.jinja", "r") as f:
    src_template = Template(f.read())


def get_heuristic_tile(
    inter_dim: int, max_num_m_blocks: int, available_tiles: list, num_cu: int
):
    empty_cu = num_cu
    tg_num = 0
    round = 0xFFFFFFFF
    selected_tile = 0

    for tile in available_tiles:
        if inter_dim % tile == 0:
            tg_num = inter_dim / tile * max_num_m_blocks
            local_round = (tg_num + num_cu - 1) / num_cu
            if local_round < round:
                round = local_round
                selected_tile = tile
                empty_cu = local_round * num_cu - tg_num
            elif local_round == round:
                if empty_cu > (local_round * num_cu - tg_num):
                    round = local_round
                    selected_tile = tile
                    empty_cu = local_round * num_cu - tg_num
    return selected_tile


def select_tile(
    input_dtype: str,
    gate_fusion: str,
    gate_dtype: str,
    inter_dim: int,
    max_num_m_blocks: int,
    num_cu: int,
    a16: bool = False,
    fc_scale_blkn: int = None,
    fc_scale_blkk: int = None,
):
    if a16:
        if gate_fusion == "g1u1":
            if gate_dtype == "uint8_t":
                if inter_dim % 320 == 0:
                    return 320
                else:
                    raise ValueError(
                        f"Unsupported inter_dim {inter_dim}, which should be divisible by 320"
                    )
            elif gate_dtype == "__hip_fp8_e4m3_fnuz":
                selected_tile = get_heuristic_tile(
                    inter_dim, max_num_m_blocks, [512, 320], num_cu
                )
                if selected_tile == 0:
                    raise ValueError(
                        f"Unsupported inter_dim {inter_dim}, which should be divisible by 320 or 512"
                    )
                return selected_tile
    else:
        if gate_fusion == "g1u0":
            tiles = [512, 448, 384, 320, 256, 192, 128]
            selected_tile = 0
            for tile in tiles:
                if inter_dim % tile == 0:
                    selected_tile = tile
                    break
            if selected_tile == 0:
                raise ValueError(
                    f"Unsupported inter_dim {inter_dim}, which should be divisible by 128, 192, 256, 320, 384, 448 or 512"
                )
            return selected_tile
        elif gate_fusion == "g1u1":
            if gate_dtype in ["uint32_t", "int32_t"]:
                selected_tile = get_heuristic_tile(
                    inter_dim, max_num_m_blocks, [512, 256, 128], num_cu
                )
                if selected_tile == 0:
                    raise ValueError(
                        f"Unsupported inter_dim {inter_dim}, which should be divisible by 128, 256 or 512"
                    )
                return selected_tile
            elif input_dtype == "uint8_t":
                selected_tile = get_heuristic_tile(
                    inter_dim,
                    max_num_m_blocks,
                    [512, 448, 384, 320, 256, 192, 128],
                    num_cu,
                )
                if selected_tile == 0:
                    raise ValueError(
                        f"Unsupported inter_dim {inter_dim}, which should be divisible by 128, 192, 256, 320, 384, 448 or 512"
                    )
                return selected_tile
            elif input_dtype == "__hip_fp8_e4m3_fnuz":
                selected_tile = get_heuristic_tile(
                    inter_dim,
                    max_num_m_blocks,
                    [512, 448, 384, 320, 256, 192, 128],
                    num_cu,
                )
                if selected_tile == 0:
                    raise ValueError(
                        f"Unsupported inter_dim {inter_dim}, which should be divisible by 128, 192, 256, 320, 384, 448 or 512"
                    )
                return selected_tile
            elif input_dtype == "__hip_bfloat16":
                if (
                    inter_dim % 256 == 0
                    and fc_scale_blkn == 128
                    and fc_scale_blkk == 128
                ):
                    return 256
                else:
                    raise ValueError(
                        f"Unsupported inter_dim {inter_dim}, which should be divisible by 256 and fc_scale_blkn and fc_scale_blkk should be 128"
                    )
    return 512


def select_hsaco(
    input_dtype: str,
    gate_fusion: str,
    gate_dtype: str,
    activation: str,
    selected_tile: int,
    is_smooth_scale=False,
    a16=False,
    enable_vskip=False,
):
    if a16:
        if gate_fusion == "g1u1":
            if gate_dtype == "uint8_t":
                return (
                    f"fmoe_int8_g1u1_smf_subGU_{selected_tile}",
                    f"fmoe_int8_g1u1_smf_subGU_{selected_tile}.co",
                    "uint8_t",
                    "uint16_t",
                    "true",
                )
            elif gate_dtype == "__hip_fp8_e4m3_fnuz":
                return (
                    f"fmoe_fp8_g1u1_smf_subGU_{selected_tile}",
                    f"fmoe_fp8_g1u1_smf_subGU_{selected_tile}.co",
                    "uint8_t",
                    "uint16_t",
                    "true",
                )
        elif gate_fusion == "g1u0":
            return (
                "fmoe_kernel_func",
                "fmoe_int8_g1u0_smf.co",
                "uint8_t",
                "uint16_t",
                "true",
            )
    else:
        if gate_fusion == "g1u0":
            if input_dtype == "__half":
                return (
                    "fmoe_kernel_func",
                    "fmoe_f16.co",
                    "uint16_t",
                    "uint16_t",
                    "false",
                )
            elif input_dtype == "__hip_bfloat16":
                return (
                    "fmoe_kernel_func",
                    "fmoe_b16.co",
                    "uint16_t",
                    "uint16_t",
                    "false",
                )
            elif input_dtype == "uint8_t":
                if activation == "gelu":
                    return (
                        f"fmoe_int8_g1u0_subGU_{selected_tile}_gelu",
                        f"moe/gelu/fmoe_int8_g1u0_subGU_{selected_tile}_gelu.co",
                        "uint8_t",
                        "uint16_t",
                        "false",
                    )
                elif activation == "silu":
                    return (
                        f"fmoe_int8_g1u0_subGU_{selected_tile}",
                        f"fmoe/silu/fmoe_int8_g1u0_subGU_{selected_tile}.co",
                        "uint8_t",
                        "uint16_t",
                        "false",
                    )

        elif gate_fusion == "g1u1":
            if gate_dtype in ["uint32_t", "int32_t"]:
                return (
                    f"fmoe_int4fp8_g1u1_subGU_{selected_tile}",
                    f"fmoe_int4fp8_g1u1_subGU_{selected_tile}_gelu.co",
                    "uint8_t",
                    "uint16_t",
                    "false",
                )
            elif input_dtype == "uint8_t":
                if is_smooth_scale:
                    return (
                        f"fmoe_int8_g1u1_multix_subGU_{selected_tile}",
                        f"fmoe_int8_g1u1_multix_subGU_{selected_tile}.co",
                        "uint8_t",
                        "uint16_t",
                        "false",
                    )
                elif activation == "gelu":
                    return (
                        f"fmoe_int8_g1u1_subGU_{selected_tile}_gelu",
                        f"fmoe/gelu/fmoe_int8_g1u1_subGU_{selected_tile}_gelu.co",
                        "uint8_t",
                        "uint16_t",
                        "false",
                    )
                elif activation == "silu":
                    return (
                        f"fmoe_int8_g1u1_subGU_{selected_tile}",
                        f"fmoe/silu/fmoe_int8_g1u1_subGU_{selected_tile}.co",
                        "uint8_t",
                        "uint16_t",
                        "false",
                    )
            elif input_dtype == "__hip_fp8_e4m3_fnuz":
                if is_smooth_scale:
                    return (
                        f"fmoe_fp8_g1u1_multix_subGU_{selected_tile}",
                        f"fmoe_fp8_g1u1_multix_subGU_{selected_tile}.co",
                        "uint8_t",
                        "uint16_t",
                        "false",
                    )
                elif activation == "gelu":
                    return (
                        f"fmoe_fp8_g1u1_subGU_{selected_tile}_gelu",
                        f"fmoe/gelu/fmoe_fp8_g1u1_subGU_{selected_tile}_gelu.co",
                        "uint8_t",
                        "uint16_t",
                        "false",
                    )
                elif activation == "silu":
                    return (
                        f"fmoe_fp8_g1u1_subGU_{selected_tile}",
                        f"fmoe/silu/fmoe_fp8_g1u1_subGU_{selected_tile}.co",
                        "uint8_t",
                        "uint16_t",
                        "false",
                    )
            elif input_dtype == "__hip_bfloat16":
                if enable_vskip:
                    return (
                        f"fmoe_fp8_blockscale_g1u1_subGU_{selected_tile}",
                        f"fmoe_fp8_blockscale_g1u1_subGU_{selected_tile}.co",
                        "uint8_t",
                        "uint16_t",
                        "false",
                    )
                else:
                    return (
                        f"fmoe_fp8_blockscale_g1u1_novs_subGU_{selected_tile}",
                        f"fmoe_fp8_blockscale_g1u1_novs_subGU_{selected_tile}.co",
                        "uint8_t",
                        "uint16_t",
                        "false",
                    )

    raise ValueError("Unsupported condition")


def compile(
    input_dtype: str,
    gate_fusion: str,
    gate_dtype: str,
    activation: str,
    selected_tile: int,
    is_smooth_scale=False,
    a16=False,
    enable_vskip=False,
    block_size=32,
    func_name=None,
    folder=None,
):
    if func_name is None:
        func_name = get_default_func_name(
            MD_NAME,
            (
                input_dtype,
                gate_fusion,
                gate_dtype,
                activation,
                selected_tile,
                is_smooth_scale,
                a16,
                enable_vskip,
                block_size,
            ),
        )

    if folder is None:
        folder = func_name

    if not_built(folder):
        kernel_name, co_name, input_dtype, output_dtype, switch_gxy = select_hsaco(
            input_dtype,
            gate_fusion,
            gate_dtype,
            activation,
            selected_tile,
            is_smooth_scale,
            a16,
            enable_vskip,
        )
        bin_size, bin_data = transfer_hsaco(f"{AITER_CORE_DIR}/hsa/{co_name}")
        return compile_template_op(
            src_template,
            MD_NAME,
            [
                f"{AITER_CORE_DIR}/csrc/cpp_itfs/utils.h",
                f"{AITER_CORE_DIR}/csrc/include",
                f"{CK_DIR}/include",
                f"{CK_DIR}/example/ck_tile/13_moe_sorting/moe_sorting_api.hpp",
            ],
            [f"{CK_DIR}/example/ck_tile/13_moe_sorting/moe_sorting_api.cpp"],
            bin_size=bin_size,
            bin_data=bin_data,
            input_dtype=input_dtype,
            output_dtype=output_dtype,
            kernel_name=kernel_name,
            selected_tile=selected_tile,
            switch_gxy=switch_gxy,
            block_size=block_size,
            func_name=func_name,
            folder=folder,
        )
    else:
        return run_lib(func_name, folder)


def asm_moe(
    hidden_states,  # [num_tokens, dim] M,K
    w1,  # [expert(local_expert:EP), inter_dim*2, dim] N,K
    w2,  # [expert(local_expert:EP), dim, inter_dim]
    topk_weight,
    topk_ids,
    # following for int8 quant
    fc1_scale=None,  # [expert(local_expert:EP), inter_dim, 1]
    fc2_scale=None,  # [expert(local_expert:EP), model_dim, 1]
    fc1_smooth_scale=None,  # [expert(local_expert:EP), 1, model_dim]
    fc2_smooth_scale=None,  # [expert(local_expert:EP), 1, inter_dim]
    a16=False,
    per_tensor_quant_scale=None,
    expert_mask=None,
    activation="silu",
    output=None,  # [num_tokens, dim]
    sorted_token_ids=None,
    sorted_weights=None,
    sorted_expert_ids=None,
    num_valid_ids=None,
    a8=None,
    a8_scale=None,
    workspace=None,
):
    import torch
    from csrc.cpp_itfs.torch_utils import torch_to_hip_types, torch_to_c_types

    E, dim, inter_dim = w2.shape
    global_E = E
    if expert_mask is not None:
        global_E = expert_mask.numel()
    num_tokens, topk = topk_ids.shape
    device = hidden_states.device
    lastdim_mul = 8 if w1.dtype in {torch.int32, torch.uint32} else 1
    if fc1_smooth_scale is None:
        gate_fusion = "g1u0"
    elif a16:
        if (
            w1.dtype in [torch.float8_e4m3_fnuz, torch.int8]
            and inter_dim * 2 == w1.shape[1]
        ):
            gate_fusion = "g1u1"
        elif w1.dtype == torch.int8 and inter_dim == w1.shape[1]:
            gate_fusion = "g1u0"
        else:
            raise ValueError(
                f"Unsupported w1.dtype {w1.dtype}, which should be {torch.float8_e4m3_fnuz} or {torch.int8} and w1.shape[1] should be {inter_dim} or {inter_dim*2}"
            )
    else:
        if fc1_smooth_scale is not None:
            if expert_mask is not None:
                local_expert_hash = expert_mask.cumsum(0, dtype=torch.int32)
                local_expert_hash[local_expert_hash > 0] -= 1

        if inter_dim * lastdim_mul == w1.shape[1]:
            gate_fusion = "g1u0"
        elif inter_dim * 2 * lastdim_mul == w1.shape[1]:
            gate_fusion = "g1u1"
        else:
            raise ValueError(
                f"Invalid MoE weight: {w1.shape=} {w2.shape=} {lastdim_mul}"
            )

    input_dtype, gate_dtype = torch_to_hip_types(hidden_states.dtype, w1.dtype)

    block_size = 32
    if sorted_token_ids is None:
        max_num_tokens_padded = topk_ids.numel() + global_E * block_size - topk
        sorted_token_ids = torch.empty(
            (max_num_tokens_padded,), dtype=torch.int32, device=device
        )
    else:
        max_num_tokens_padded = sorted_token_ids.shape[0]

    if sorted_weights is None:
        sorted_weights = torch.empty(
            (max_num_tokens_padded,), dtype=torch.float, device=device
        )
    if sorted_expert_ids is None:
        max_num_m_blocks = int((max_num_tokens_padded + block_size - 1) // block_size)
        sorted_expert_ids = torch.empty(
            (max_num_m_blocks,), dtype=torch.int32, device=device
        )
    else:
        max_num_m_blocks = sorted_expert_ids.shape[0]

    if num_valid_ids is None:
        num_valid_ids = torch.empty((1,), dtype=torch.int32, device=device)

    if output is None:
        output = torch.empty(
            (num_tokens, dim), dtype=hidden_states.dtype, device=device
        )

    num_cu = torch.cuda.get_device_properties().multi_processor_count
    selected_tile = select_tile(
        input_dtype, gate_fusion, gate_dtype, inter_dim, max_num_m_blocks, num_cu, a16
    )
    func = compile(
        input_dtype,
        gate_fusion,
        gate_dtype,
        activation,
        selected_tile,
        bool(fc2_smooth_scale),
        a16,
        block_size=block_size,
    )
    workspace_size_func = run_lib("moe_sorting_get_workspace_size", func.__name__)
    workspace_size = ctypes.c_int(0)
    workspace_size_func(
        *torch_to_c_types(num_tokens, global_E), ctypes.pointer(workspace_size)
    )
    workspace_size = workspace_size.value
    if workspace_size > 0 and workspace is None:
        workspace = torch.zeros(workspace_size, dtype=topk_ids.dtype, device=device)
    if fc2_smooth_scale is None:
        func(
            *torch_to_c_types(
                output,
                hidden_states,
                w1,
                w2,
                topk_weight,
                topk_ids,
                sorted_token_ids,
                sorted_weights,
                sorted_expert_ids,
                num_valid_ids,
                num_tokens,
                dim,
                inter_dim,
                topk,
                global_E,
                max_num_m_blocks,
                output.nbytes,
                hidden_states.stride(0),
                expert_mask,
                workspace,
                torch.cuda.current_stream(),
            )
        )
    else:
        func(
            *torch_to_c_types(
                output,
                hidden_states,
                w1,
                w2,
                topk_weight,
                topk_ids,
                fc1_scale,
                fc2_scale,
                fc1_smooth_scale,
                fc2_smooth_scale,
                a16,
                per_tensor_quant_scale,
                expert_mask,
                activation,
            )
        )
    return output


if __name__ == "__main__":
    pass
